import logging

from boto.exception import BotoServerError
from boto import utils
from tornado import testing
import mock

from kingpin.actors.aws import settings
from kingpin.actors import exceptions
from kingpin.actors.aws import base

log = logging.getLogger(__name__)


class TestBase(testing.AsyncTestCase):

    def setUp(self):
        super(TestBase, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        reload(base)

    def test_region_check(self):
        with self.assertRaises(exceptions.InvalidOptions):
            base.AWSBaseActor('Unit Test Action', {'region': 'fail'})

    def test_zone_check(self):
        actor = base.AWSBaseActor('Unit Test Action',
                                  {'region': 'us-west-1d'})
        self.assertEquals(actor.ec2_conn.region.name, 'us-west-1')

    @testing.gen_test
    def test_thread_exception(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers = mock.MagicMock()
        exc = BotoServerError(403, 'The security token')
        actor.elb_conn.get_all_load_balancers.side_effect = exc

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor._find_elb('')

    @testing.gen_test
    def test_find_elb(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers.return_value = ['test']

        elb = yield actor._find_elb('')

        self.assertEquals(elb, 'test')
        self.assertEquals(actor.elb_conn.get_all_load_balancers.call_count, 1)

        actor.elb_conn.get_all_load_balancers.assert_called_with(
            load_balancer_names='')

    @testing.gen_test
    def test_find_elb_error(self):
        actor = base.AWSBaseActor('Unit Test Action', {})

        # Pretend the request worked, but there are no ELBs
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers = mock.Mock(return_value=[])
        with self.assertRaises(base.ELBNotFound):
            yield actor._find_elb('')

    @testing.gen_test
    def test_find_elb_exception_error(self):
        actor = base.AWSBaseActor('Unit Test Action', {})

        # Pretend the request worked, but there are no ELBs
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers = mock.MagicMock()
        actor.elb_conn.get_all_load_balancers.side_effect = BotoServerError(
            400, 'LoadBalancerNotFound')
        with self.assertRaises(base.ELBNotFound):
            yield actor._find_elb('')

        # Pretend the request worked, but there are no ELBs
        actor.elb_conn.get_all_load_balancers.side_effect = BotoServerError(
            401, 'SomeOtherError')
        with self.assertRaises(BotoServerError):
            yield actor._find_elb('')

    @testing.gen_test
    def test_get_meta_data(self):
        actor = base.AWSBaseActor('Unit Test Action', {})

        with mock.patch.object(utils, 'get_instance_metadata') as md:
            md.return_value = {'ut-key': 'ut-value'}
            meta = yield actor._get_meta_data('ut-key')

        self.assertEquals(meta, 'ut-value')

    @testing.gen_test
    def test_get_meta_data_error(self):
        actor = base.AWSBaseActor('Unit Test Action', {})

        with mock.patch.object(utils, 'get_instance_metadata') as md:
            md.return_value = {}
            with self.assertRaises(base.InvalidMetaData):
                yield actor._get_meta_data('ut-key')

        with mock.patch.object(utils, 'get_instance_metadata') as md:
            md.return_value = {'key': 'value'}
            with self.assertRaises(base.InvalidMetaData):
                yield actor._get_meta_data('ut-key')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2014 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.base`
^^^^^^^^^^^^^^^^^^^^^^^^^^

Base Actor object class

An Actor object is a class that executes a single logical action
on a resource as part of your deployment structure. For example, you
may have an Actor that launches a server array in RightScale, or you
may have one that sends an email.

Each Actor object should do one thing, and one thing only. Its responsible
for being able to execute the operation in both 'dry' and 'non-dry' modes.

The behavior for 'dry' mode can contain real API calls, but should not make
any live changes. It is up to the developer of the Actor to define what
'dry' mode looks like for that particular action.
"""

import json
import logging
import os
import sys
import time

from tornado import gen
from tornado import httpclient
from tornado import httputil

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


# If super-debug logging is enabled, then we turn on the URLLIB3 HTTP
# request logging. This is extremely verbose and insecure, but useful
# for troubleshooting. URLLIB3 is used by several actors (aws, rightscale),
# so we do this setup here in the base actor class.
if os.getenv('URLLIB_DEBUG', None):
    utils.super_httplib_debug_logging()

# Allow the user to override the default_timeout for all actors by setting an
# environment variable
DEFAULT_TIMEOUT = os.getenv('DEFAULT_TIMEOUT', 3600)


class LogAdapter(logging.LoggerAdapter):

    def process(self, msg, kwargs):
        return ('[%s%s] %s' % (self.extra['dry'], self.extra['desc'], msg),
                kwargs)


class BaseActor(object):

    """Abstract base class for Actor objects."""

    # {
    #     'option_name': (type, default, "Long description of the option"),
    # }
    #
    # If `default` is REQUIRED then the option requires user specified input
    #
    # Example:
    # {
    #    'room': (str, REQUIRED, 'Hipchat room to notify'),
    #    'from': (str, 'Kingpin', 'User that sends the message')
    # }
    all_options = {}

    # Set the default timeout for the gen.with_timeout() wrapper that we use to
    # monitor and control the length of execution of a single Actor.
    default_timeout = DEFAULT_TIMEOUT

    # Context separators. These define the left-and-right identifiers of a
    # 'contextual token' in the actor. By default this is { and }, so a
    # contextual token looks like '{KEY}'.
    left_context_separator = '{'
    right_context_separator = '}'

    # Ensure that at __init__ time, if the self._options dict is not completely
    # filled in properly (meaning there are no left-over {KEY}'s), we throw an
    # exception. This will change in the future when we have some concept of a
    # second 'global runtime context object'.
    strict_init_context = True

    def __init__(self, desc, options, dry=False, warn_on_failure=False,
                 condition=True, init_context={}, timeout=None):
        """Initializes the Actor.

        Args:
            desc: (Str) description of the action being executed.
            options: (Dict) Key/Value pairs that have the options
                     for this action. Values should be primitives.
            dry: (Bool) or not this Actor will actually make changes.
            warn_on_failure: (Bool) Whether this actor ignores its return
                             value and always succeeds (but warns).
            condition: (Bool) Whether to run this actor.
            init_context: (Dict) Key/Value pairs used at instantiation
                time to replace {KEY} strings in the actor definition.
                This is usually driven by the group.Sync/Async actors.
            timeout: (Str/Int/Float) Timeout in seconds for the actor.
        """
        self._type = '%s.%s' % (self.__module__, self.__class__.__name__)
        self._desc = desc
        self._options = options
        self._dry = dry
        self._warn_on_failure = warn_on_failure
        self._condition = condition
        self._init_context = init_context

        self._timeout = timeout
        if timeout is None:
            self._timeout = self.default_timeout

        # strict about this -- but in the future, when we have a
        # runtime_context object, we may loosen this restriction).
        self._fill_in_contexts(context=self._init_context,
                               strict=self.strict_init_context)

        self._setup_log()
        self._setup_defaults()
        self._validate_options()  # Relies on _setup_log() above

        # Fill in any options with the supplied initialization context. Be
        self.log.debug('Initialized (warn_on_failure=%s, '
                       'strict_init_context=%s)' %
                       (warn_on_failure, self.strict_init_context))

    def _setup_log(self):
        """Create a customized logging object based on the LogAdapter."""
        name = '%s.%s' % (self.__module__, self.__class__.__name__)
        logger = logging.getLogger(name)
        dry_str = 'DRY: ' if self._dry else ''

        self.log = LogAdapter(logger, {'desc': self._desc, 'dry': dry_str})

    def _setup_defaults(self):
        """Populate options with defaults if they aren't set."""

        for option, definition in self.all_options.items():
            if option not in self._options:
                default = definition[1]
                if default is not REQUIRED:
                    self._options.update({option: default})

    def _validate_options(self):
        """Validate that all the required options were passed in.

        Args:
            options: A dictionary of options.

        Raises:
            exceptions.InvalidOptions
        """

        # Loop through all_options, and find the required ones
        required = [opt_name
                    for (opt_name, definition) in self.all_options.items()
                    if definition[1] is REQUIRED]

        self.log.debug('Checking for required options: %s' % required)
        option_errors = []
        option_warnings = []
        for opt in required:
            if opt not in self._options:
                description = self.all_options[opt][2]
                option_errors.append('Option "%s" is required: %s' % (
                                     opt, description))

        for opt, value in self._options.items():
            if opt not in self.all_options:
                option_warnings.append('Option "%s" is not expected.' % opt)
                continue

            expected_type = self.all_options[opt][0]

            # Unicode is not a `str` but it is a `basestring`
            # Cast the passed value explicitly as a string
            if isinstance(value, basestring):
                value = str(value)

            if not (value is None or isinstance(value, expected_type)):
                message = 'Option "%s" has to be %s and is %s.' % (
                    opt, expected_type, type(value))
                option_errors.append(message)

        for w in option_warnings:
            self.log.warning(w)

        if option_errors:
            for e in option_errors:
                self.log.critical(e)
            raise exceptions.InvalidOptions(
                'Found %s issue(s) with passed options.' % len(option_errors))

    def option(self, name):
        """Return the value for a given Actor option."""

        return self._options.get(name)

    def readfile(self, path):
        """Return file contents as a string.

        Raises:
            InvalidOptions if file is not found, or readable.
        """

        try:
            with open(path) as f:
                contents = f.read()
        except IOError as e:
            raise exceptions.InvalidOptions(e)

        return contents

    def timer(f):
        """Coroutine-compatible function timer.

        Records statistics about how long a given function took, and logs them
        out in debug statements. Used primarily for tracking Actor execute()
        methods, but can be used elsewhere as well.

        Example usage:
            >>> @gen.coroutine
            ... @timer()
            ... def execute(self):
            ...     raise gen.Return()
        """

        def _wrap_in_timer(self, *args, **kwargs):
            # Log the start time
            start_time = time.time()

            # Begin the execution
            ret = yield gen.coroutine(f)(self, *args, **kwargs)

            # Log the finished execution time
            exec_time = "%.2f" % (time.time() - start_time)
            self.log.debug('%s.%s() execution time: %ss' %
                           (self._type, f.__name__, exec_time))

            raise gen.Return(ret)
        return _wrap_in_timer

    @gen.coroutine
    def timeout(self, f, *args, **kwargs):
        """Wraps a Coroutine method in a timeout.

        Used to wrap the self.execute() method in a timeout that will raise an
        ActorTimedOut exception if an actor takes too long to execute.

        *Note, Tornado 4+ does not allow you to actually kill a task on the
        IOLoop.*  This means that all we are doing here is notifying the caller
        (through the raised exception) that a problem has happened.

        Fairly simple Actors should actually 'stop executing' when this
        exception is raised. Complex actors with very unique behaviors though
        (like the rightsacle.server_array.Execute actor) have the ability to
        continue to execute in the background until the Kingpin application
        quits. It is not the job of this method to try to kill these actors,
        but just to let the user know that a failure has happened.
        """

        # Get our timeout setting, or fallback to the default
        self.log.debug('%s.%s() deadline: %s(s)' %
                       (self._type, f.__name__, self._timeout))

        # Get our Future object but don't yield on it yet, This starts the
        # execution, but allows us to wrap it below with the
        # 'gen.with_timeout' function.
        fut = f(*args, **kwargs)

        # If no timeout is set (none, or 0), then we just yield the Future and
        # return its results.
        if not self._timeout:
            ret = yield fut
            raise gen.Return(ret)

        # Generate a timestamp in the future at which point we will raise
        # an alarm if the actor is still executing
        deadline = time.time() + float(self._timeout)

        # Now we yield on the gen_with_timeout function
        try:
            ret = yield gen.with_timeout(
                deadline, fut, quiet_exceptions=(exceptions.ActorTimedOut))
        except gen.TimeoutError:
            msg = ('%s.%s() execution exceeded deadline: %ss' %
                   (self._type, f.__name__, self._timeout))
            self.log.error(msg)
            raise exceptions.ActorTimedOut(msg)

        raise gen.Return(ret)

    def _check_condition(self):
        """Check if specified condition allows this actor to run.

        Evaluate self._condition to figure out if this actor should run.
        The only exception to simply casting this variable to bool is if
        the value of self._condition is a string "False" or string "0".
        """

        try:  # Treat as string
            value = self._condition.lower()
            check = (value not in ('false', '0'))
        except AttributeError:  # Not a string
            value = self._condition
            check = bool(value)

        return check

    def _fill_in_contexts(self, context={}, strict=True):
        """Parses self._options and updates it with the supplied context.

        Parses the objects self._options dict (by converting it into a JSON
        string, substituting, and then turning it back into a dict) and the
        self._desc string and replaces any {KEY}s with the valoues from the
        context dict that was supplied.

        Args:
            strict: bool whether or not to allow missing context keys to be
                    skipped over.

        Raises:
            exceptions.InvalidOptions
        """
        try:
            self._desc = utils.populate_with_tokens(
                self._desc,
                context,
                self.left_context_separator,
                self.right_context_separator,
                strict=strict)
        except LookupError as e:
            msg = 'Context for description failed: %s' % e
            raise exceptions.InvalidOptions(msg)

        # Convert our self._options dict into a string for fast parsing
        options_string = json.dumps(self._options)

        # Generate a new string with the values parsed out. At this point, if
        # any value is un-matched, an exception is raised and execution fails.
        # This stops execution during a dry run, before any live changes are
        # made.
        try:
            new_options_string = utils.populate_with_tokens(
                options_string,
                context,
                self.left_context_separator,
                self.right_context_separator,
                strict=strict)
        except LookupError as e:
            msg = 'Context for options failed: %s' % e
            raise exceptions.InvalidOptions(msg)

        # Finally, convert the string back into a dict and store it.
        self._options = json.loads(new_options_string)

    @gen.coroutine
    @timer
    def execute(self):
        """Executes an actor and yields the results when its finished.

        Calls an actors private _execute() method and either returns the result
        (through gen.Return) or handles any exceptions that are raised.

        RecoverableActorFailure exceptions are potentially swallowed up (and
        warned) if the self._warn_on_failure flag is set. Otherwise, they're
        logged and re-raised. All other ActorException exceptions are caught,
        logged and re-raised.

        We have a generic catch-all exception handling block as well, because
        third party Actor classes may or may not catch all appropriate
        exceptions. This block is mainly here to prevent the entire app from
        failing due to a poorly written Actor.

        Raises:
            gen.Return(result)
        """
        self.log.debug('Beginning')

        # Any exception thats raised by an actors _execute() method will
        # automatically cause actor failure and we return right away.
        result = None

        if not self._check_condition():
            self.log.warning('Skipping execution. Condition: %s' %
                             self._condition)
            raise gen.Return()

        try:
            result = yield self.timeout(self._execute)
        except exceptions.ActorException as e:
            # If exception is not RecoverableActorFailure
            # or if warn_on_failure is not set, then escalate.
            recover = isinstance(e, exceptions.RecoverableActorFailure)
            if not recover or not self._warn_on_failure:
                self.log.critical(e)
                raise

            # Otherwise - flag this failure as a warning, and continue
            self.log.warning(e)
            self.log.warning(
                'Continuing execution even though a failure was '
                'detected (warn_on_failure=%s)' % self._warn_on_failure)
        except Exception as e:
            # We don't like general exception catch clauses like this, but
            # because actors can be written by third parties and automatically
            # imported, its impossible for us to catch every exception
            # possible. This is a failsafe thats meant to throw a strong
            # warning.
            log.critical('Unexpected exception caught! '
                         'Please contact the author (%s) and provide them '
                         'with this stacktrace' %
                         sys.modules[self.__module__].__author__)
            self.log.exception(e)
            raise exceptions.ActorException(e)
        else:
            self.log.debug('Finished successfully, return value: %s' % result)

        # If we got here, we're exiting the actor cleanly and moving on.
        raise gen.Return(result)


class HTTPBaseActor(BaseActor):

    """Abstract base class for an HTTP-client based Actor object.

    This class provides common methods for getting access to asynchronous
    HTTP clients, wrapping the executions in appropriate try/except blocks,
    timeouts, etc.

    If you're writing an Actor that uses a remote REST API, this is the
    base class you should subclass from.
    """

    headers = None

    def _get_http_client(self):
        """Returns an asynchronous web client object

        The object is actually of type SimpleAsyncHTTPClient
        """
        return httpclient.AsyncHTTPClient()

    def _get_method(self, post):
        """Returns the appropriate HTTP Method based on the supplied Post data.

        Args:
            post: The post body you intend to submit in the URL request

        Returns:
            'GET' or 'POST'
        """
        # If there is no post data, set the request method to GET
        if post is not None:
            return 'POST'
        else:
            return 'GET'

    def _generate_escaped_url(self, url, args):
        """Takes in a dictionary of arguments and returns a URL line.

        Sorts the arguments so that the returned string is predictable and in
        alphabetical order. Effectively wraps the tornado.httputil.url_concat
        method and properly strips out None values, as well as lowercases
        Bool values.

        Args:
            url: (Str) The URL to append the arguments to
            args: (Dict) Key/Value arguments. Values should be primitives.

        Returns:
            A URL encoded string like this: <url>?foo=bar&abc=xyz
        """

        # Remove keys from the arguments where the value is None
        args = dict((k, v) for k, v in args.iteritems() if v)

        # Convert all Bool values to lowercase strings
        for key, value in args.iteritems():
            if type(value) is bool:
                args[key] = str(value).lower()

        # Now generate the URL
        full_url = httputil.url_concat(url, sorted(args.items()))
        self.log.debug('Generated URL: %s' % full_url)

        return full_url

    # TODO: Add a retry/backoff timer here. If the remote endpoint returns
    # garbled data (ie, maybe a 500 errror or something else thats not in
    # JSON format, we should back off and try again.
    @gen.coroutine
    def _fetch(self, url, post=None, auth_username=None, auth_password=None):
        """Executes a web request asynchronously and yields the body.

        Args:
            url: (Str) The full url path of the API call
            post: (Str) POST body data to submit (if any)
            auth_username: (str) HTTP auth username
            auth_password: (str) HTTP auth password
        """

        # Generate the full request URL and log out what we're doing...
        self.log.debug('Making HTTP request to %s with data: %s' % (url, post))

        # Create the http_request object
        http_client = self._get_http_client()
        http_request = httpclient.HTTPRequest(
            url=url,
            method=self._get_method(post),
            body=post,
            headers=self.headers,
            auth_username=auth_username,
            auth_password=auth_password,
            follow_redirects=True,
            max_redirects=10)

        # Execute the request and raise any exception. Exceptions are not
        # caught here because they are unique to the API endpoints, and thus
        # should be handled by the individual Actor that called this method.
        http_response = yield http_client.fetch(http_request)

        try:
            body = json.loads(http_response.body)
        except ValueError as e:
            raise exceptions.UnparseableResponseFromEndpoint(
                'Unable to parse response from remote API as JSON: %s' % e)

        # Receive a successful return
        raise gen.Return(body)



"""Simple integration tests for the AWS Base."""

from nose.plugins.attrib import attr
import logging

from tornado import testing

from kingpin.actors import exceptions
from kingpin.actors.aws import base
from kingpin.actors.aws import settings

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'

log = logging.getLogger(__name__)
logging.getLogger('boto').setLevel(logging.INFO)


class IntegrationBase(testing.AsyncTestCase):
    """High level AWS Base testing."""

    integration = True

    region = 'us-east-1'
    elb_name = 'kingpin-integration-test'

    @attr('integration')
    @testing.gen_test(timeout=60)
    def integration_01a_check_credentials(self):

        settings.AWS_ACCESS_KEY_ID = 'fake'
        settings.AWS_SECRET_ACCESS_KEY = 'fake'
        actor = base.AWSBaseActor('Test', {'region': self.region})

        # Executing a random function call that is wrapped in _retry.
        # Credentials should fail before "ELB not found" should be raised.
        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor._find_elb('unit-test')

        reload(settings)

    @attr('integration')
    @testing.gen_test(timeout=60)
    def integration_02a_find_elb(self):

        actor = base.AWSBaseActor('Test', {'region': self.region})

        LB = yield actor._find_elb(self.elb_name)

        self.assertEquals(LB.name, self.elb_name)



import logging

from boto.exception import BotoServerError
from tornado import testing
import mock

from kingpin.actors import exceptions
from kingpin.actors.aws import settings
from kingpin.actors.aws import iam

log = logging.getLogger(__name__)


class TestUploadCert(testing.AsyncTestCase):

    def setUp(self):
        super(TestUploadCert, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        reload(iam)

    @testing.gen_test
    def test_execute(self):
        actor = iam.UploadCert(
            'Unit Test',
            {'name': 'test',
             'public_key_path': 'test',
             'private_key_path': 'test',
             'cert_chain_path': 'test'}
        )
        actor.iam_conn = mock.Mock()

        actor.readfile = mock.Mock()
        actor.iam_conn.upload_server_cert.side_effect = [
            None
        ]
        yield actor._execute()

        # call count is 1 -- one extra retry due to BotoServerError above.
        self.assertEquals(actor.iam_conn.upload_server_cert.call_count, 1)
        actor.iam_conn.upload_server_cert.assert_called_with(
            path=None,
            private_key=actor.readfile(),
            cert_body=actor.readfile(),
            cert_name='test',
            cert_chain=actor.readfile())

    @testing.gen_test
    def test_execute_dry(self):
        actor = iam.UploadCert(
            'Unit Test',
            {'name': 'test',
             'public_key_path': 'test',
             'private_key_path': 'test',
             'cert_chain_path': 'test'},
            dry=True
        )
        actor.iam_conn = mock.Mock()

        actor.readfile = mock.Mock()
        open_patcher = mock.patch('%s.open' % actor.__module__,
                                  create=True)
        with open_patcher:
            yield actor._execute()

        self.assertEquals(actor.iam_conn.upload_server_cert.call_count, 0)


class TestDeleteCert(testing.AsyncTestCase):

    def setUp(self):
        super(TestDeleteCert, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        reload(iam)

    @testing.gen_test(timeout=60)
    def test_delete_cert_dry(self):
        actor = iam.DeleteCert('Test', {'name': 'test'}, dry=True)
        actor.iam_conn = mock.Mock()

        yield actor.execute()

        actor.iam_conn.get_server_certificate.assert_called_with('test')
        self.assertEquals(actor.iam_conn.get_server_certificate.call_count, 1)

        err = BotoServerError('400', 'Broken!')
        actor.iam_conn.get_server_certificate.side_effect = err

        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield actor.execute()

    @testing.gen_test(timeout=60)
    def test_delete_cert(self):
        actor = iam.DeleteCert(
            'Test',
            {'name': 'test'})
        actor.iam_conn = mock.Mock()

        yield actor.execute()



import sys
import time
from .rightscale import RightScale as _RS
from .util import find_by_name


__all__ = [
    'get_api',
    'get_accounts',
    'list_instances',
    'run_script_on_server',
    'get_by_path',
    ]


_api = None


def get_api():
    global _api
    if not _api:
        _api = _RS()
    return _api


def get_accounts():
    """
    Returns the RightScale accounts for the given login creds.
    """
    api = get_api()
    return api.sessions.accounts()


def list_instances(
        deployment_name='',
        cloud_name='EC2 us-east-1',
        view='tiny',
        ):
    """
    Returns a list of instances from your account.

    :param str deployment_name: If provided, only lists servers in the
        specified deployment.

    :param str cloud_name: The friendly name for a RightScale-supported cloud.
        E.g. ``EC2 us-east-1``, ``us-west-2``, etc...

    :param str view: The level of detail to request of RightScale.  Valid
        values are ``default``, ``extended``, ``full``, ``full_inputs_2_0``,
        ``tiny``.  Defaults to ``tiny``.

    """
    api = get_api()
    cloud = find_by_name(api.clouds, cloud_name)
    filters = ['state==operational']
    if deployment_name:
        deploy = find_by_name(api.deployments, deployment_name)
        filters.append('deployment_href==' + deploy.href)
    params = {'filter[]': filters, 'view': view}
    return cloud.instances.index(params=params)


def run_script_on_server(
        script_name,
        server_name,
        inputs=None,
        timeout_s=10,
        output=sys.stdout
        ):
    """
    Runs a RightScript and polls for status.

    Sample usage::

        from rightscale import run_script_on_server
        run_script_on_server(
                'my cool bob lol script',
                'some server',
                inputs={'BOB': 'blah blah', 'LOL': 'fubar'},
                )

    Sample output::

        status: Querying tags
        status: Querying tags
        status: Preparing execution
        status: RightScript: 'my cool bob lol script'
        status: completed: my cool bob lol script

    Defaults to printing status message to stdout, but will accept any object
    that implements ``write()`` passed in to :attr:`output`.
    """
    api = get_api()
    script = find_by_name(api.right_scripts, script_name)
    server = find_by_name(api.servers, server_name)
    path = server.links['current_instance'] + '/run_executable'

    data = {
            'right_script_href': script.href,
            }
    if inputs:
        for k, v in inputs.items():
            data['inputs[%s]' % k] = 'text:' + v
    response = api.client.post(path, data=data)
    status_path = response.headers['location']
    for i in range(timeout_s):
        status = api.client.get(status_path).json()
        summary = status.get('summary', '')
        output.write('status: %s\n' % summary)
        if summary.startswith('completed'):
            return
        time.sleep(1)
    output.write('Done waiting. Poll %s for status.\n' % status_path)


def get_by_path(path, first=False):
    """
    Search for resources using colon-separated path notation.

    E.g.::

        path = 'deployments:production:servers:haproxy'
        haproxies = get_by_path(path)

    :param bool first: Always use the first returned match for all intermediate
        searches along the path.  If this is ``False`` and an intermediate
        search returns multiple hits, an exception is raised.

    """
    api = get_api()

    cur_res = api
    parts = path.split(':')
    for part in parts:
        res = getattr(cur_res, part, None)
        if not res:
            # probably the name of the res to find
            res = find_by_name(cur_res, part)
        cur_res = res

    index = getattr(cur_res, 'index', None)
    if index:
        return index()
    return cur_res



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2014 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.cloudformation`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import logging

from boto.exception import BotoServerError
from tornado import concurrent
from tornado import gen
from tornado import ioloop

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.aws import base
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


# This executor is used by the tornado.concurrent.run_on_executor()
# decorator. We would like this to be a class variable so its shared
# across RightScale objects, but we see testing IO errors when we
# do this.
EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class CloudFormationError(exceptions.RecoverableActorFailure):

    """Raised on any generic CloudFormation error."""


class InvalidTemplate(exceptions.UnrecoverableActorFailure):

    """An invalid CloudFormation template was supplied."""


class StackAlreadyExists(exceptions.RecoverableActorFailure):

    """The requested CloudFormation stack already exists."""


class StackNotFound(exceptions.RecoverableActorFailure):

    """The requested CloudFormation stack does not exist."""


# CloudFormation has over a dozen different 'stack states'... but for the
# purposes of these actors, we really only care about a few logical states.
# Here we map the raw states into logical states.
COMPLETE = ('CREATE_COMPLETE', 'UPDATE_COMPLETE')
DELETED = ('DELETE_COMPLETE', 'ROLLBACK_COMPLETE')
IN_PROGRESS = (
    'CREATE_IN_PROGRESS', 'DELETE_IN_PROGRESS',
    'ROLLBACK_IN_PROGRESS', 'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS',
    'UPDATE_IN_PROGRESS', 'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS',
    'UPDATE_ROLLBACK_IN_PROGRESS')
FAILED = (
    'CREATE_FAILED', 'DELETE_FAILED', 'ROLLBACK_FAILED',
    'UPDATE_ROLLBACK_FAILED')


class CloudFormationBaseActor(base.AWSBaseActor):

    """Base Actor for CloudFormation tasks"""

    # Get references to existing objects that are used by the
    # tornado.concurrent.run_on_executor() decorator.
    ioloop = ioloop.IOLoop.current()

    executor = EXECUTOR

    # Used mainly for unit testing..
    all_options = {
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2')
    }

    @gen.coroutine
    def _get_stacks(self):
        """Gets a list of existing CloudFormation stacks.

        Gets a list of all of the stacks currently in the account, that are not
        in the status 'DELETE_COMPLETE'.

        Returns:
            A list of boto.cloudformation.stack.StackSummary objects.
        """
        # Get the list of all possible stack statuses from the Boto module,
        # then pull out the few that indicate a stack is no longer in
        # existence.
        self.log.debug('Getting list of stacks from Amazon..')
        statuses = list(self.cf_conn.valid_states)
        statuses.remove('DELETE_COMPLETE')
        stacks = yield self.thread(self.cf_conn.list_stacks,
                                   stack_status_filters=statuses)
        raise gen.Return(stacks)

    @gen.coroutine
    def _get_stack(self, stack):
        """Returns a cloudformation.Stack object of the requested stack.

        Args:
            stack: String name

        Returns
            <Stack Object> or <None>
        """
        stacks = yield self._get_stacks()
        self.log.debug('Checking whether stack %s exists.' % stack)
        new_list = [s for s in stacks if s.stack_name == stack]

        if len(new_list) > 0:
            raise gen.Return(new_list[0])

        raise gen.Return(None)

    @gen.coroutine
    def _wait_until_state(self, desired_states, sleep=15):
        """Indefinite loop until a stack has finished creating/deleting.

        Whether the stack has failed, suceeded or been rolled back... this
        method loops until the process has finished. If the final status is a
        failure (rollback/failed) then an exception is raised.

        Args:
            desired_states: (tuple/list) States that indicate a successful
                            operation.
            sleep: (int) Time in seconds between stack status checks

        Raises:
            StackNotFound: If the stack doesn't exist.
        """
        while True:
            stack = yield self._get_stack(self.option('name'))

            if not stack:
                msg = 'Stack "%s" not found.' % self.option('name')
                raise StackNotFound(msg)

            self.log.debug('Got stack %s status: %s' %
                           (stack.stack_name, stack.stack_status))

            # First, lets see if the stack is still in progress (either
            # creation, deletion, or rollback .. doesn't really matter)
            if stack.stack_status in IN_PROGRESS:
                self.log.info('Stack is in %s, waiting %s(s)...' %
                              (stack.stack_status, sleep))
                yield utils.tornado_sleep(sleep)
                continue

            # If the stack is in the desired state, then return
            if stack.stack_status in desired_states:
                self.log.info('Stack execution completed, final state: %s' %
                              stack.stack_status)
                raise gen.Return()

            # Lastly, if we get here, then something is very wrong and we got
            # some funky status back. Throw an exception.
            msg = 'Unxpected stack state received (%s)' % stack.stack_status
            raise CloudFormationError(msg)


class Create(CloudFormationBaseActor):

    """Creates a CloudFormation stack.

    Creates a CloudFormation stack from scratch and waits until the stack is
    fully built before exiting the actor.

    **Options**

    :capabilities:
      A list of CF capabilities to add to the stack.

    :disable_rollback:
      Set to True to disable rollback of the stack if creation failed.

    :name:
      The name of the queue to create

    :parameters:
      A dictionary of key/value pairs used to fill in the parameters for the
      CloudFormation template.

    :region:
      AWS region (or zone) string, like 'us-west-2'

    :template:
      String of path to CloudFormation template. Can either be in the form of a
      local file path (ie, `./my_template.json`) or a URI (ie
      `https://my_site.com/cf.json`).

    :timeout_in_minutes:
      The amount of time that can pass before the stack status becomes
      CREATE_FAILED.

    **Examples**

    .. code-block:: json

       { "desc": "Create production backend stack",
         "actor": "aws.cloudformation.Create",
         "options": {
           "capabilities": [ "CAPABILITY_IAM" ],
           "disable_rollback": true,
           "name": "%CF_NAME%",
           "parameters": {
             "test_param": "%TEST_PARAM_NAME%",
           },
           "region": "us-west-1",
           "template": "/examples/cloudformation_test.json",
           "timeout_in_minutes": 45,
         }
       }

    **Dry Mode**

    Validates the template, verifies that an existing stack with that name does
    not exist. Does not create the stack.
    """

    all_options = {
        'capabilities': (list, [],
                         'The list of capabilities that you want to allow '
                         'in the stack'),
        'disable_rollback': (bool, False,
                             'Set to `True` to disable rollback of the stack '
                             'if stack creation failed.'),
        'name': (str, REQUIRED, 'Name of the stack'),
        'parameters': (dict, {}, 'Parameters passed into the CF '
                                 'template execution'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2'),
        'template': (str, REQUIRED,
                     'Path to the AWS CloudFormation File. http(s)://, '
                     'file:///, absolute or relative file paths.'),
        'timeout_in_minutes': (int, 60,
                               'The amount of time that can pass before the '
                               'stack status becomes CREATE_FAILED'),
    }

    def __init__(self, *args, **kwargs):
        """Initialize our object variables."""
        super(Create, self).__init__(*args, **kwargs)

        # Check if the supplied CF template is a local file. If it is, read it
        # into memory.
        (self._template_body, self._template_url) = self._get_template_body(
            self.option('template'))

    def _get_template_body(self, template):
        """Reads in a local template file and returns the contents.

        If the template string supplied is a local file resource (has no
        URI prefix), then this method will return the contents of the file.
        Otherwise, returns None.

        Args:
            template: String with a reference to a template location.

        Returns:
            One tuple of:
              (Contents of template file, None)
              (None, URL of template)

        Raises:
            InvalidTemplate
        """
        remote_types = ('http://', 'https://')

        if template.startswith(remote_types):
            return (None, template)

        try:
            # TODO: leverage self.readfile()
            return (open(template, 'r').read(), None)
        except IOError as e:
            raise InvalidTemplate(e)

    @gen.coroutine
    def _validate_template(self):
        """Validates the CloudFormation template.

        Raises:
            InvalidTemplate
            exceptions.InvalidCredentials
        """
        if self._template_body is not None:
            self.log.info('Validating template with AWS...')
        else:
            self.log.info('Validating template (%s) with AWS...' %
                          self._template_url)

        try:
            yield self.thread(
                self.cf_conn.validate_template,
                template_body=self._template_body,
                template_url=self._template_url)
        except BotoServerError as e:
            msg = '%s: %s' % (e.error_code, e.message)

            if e.status == 400:
                raise InvalidTemplate(msg)

            raise

    @gen.coroutine
    def _create_stack(self):
        """Executes the stack creation."""
        # Create the stack, and get its ID.
        self.log.info('Creating stack %s' % self.option('name'))
        try:
            stack_id = yield self.thread(
                self.cf_conn.create_stack,
                self.option('name'),
                template_body=self._template_body,
                template_url=self._template_url,
                parameters=self.option('parameters').items(),
                disable_rollback=self.option('disable_rollback'),
                timeout_in_minutes=self.option('timeout_in_minutes'),
                capabilities=self.option('capabilities'))
        except BotoServerError as e:
            msg = '%s: %s' % (e.error_code, e.message)

            if e.status == 400:
                raise CloudFormationError(msg)

            raise

        self.log.info('Stack %s created: %s' % (self.option('name'), stack_id))
        raise gen.Return(stack_id)

    @gen.coroutine
    def _execute(self):
        stack_name = self.option('name')

        yield self._validate_template()

        # If a stack already exists, we cannot re-create it. Raise a
        # recoverable exception and let the end user decide whether this is bad
        # or not.
        exists = yield self._get_stack(stack_name)
        if exists:
            raise StackAlreadyExists('Stack %s already exists!' % stack_name)

        # If we're in dry mode, exit at this point. We can't do anything
        # further to validate that the creation process will work.
        if self._dry:
            self.log.info('Skipping CloudFormation Stack creation.')
            raise gen.Return()

        # Create the stack
        yield self._create_stack()

        # Now wait until the stack creation has finished
        yield self._wait_until_state(COMPLETE)

        raise gen.Return()


class Delete(CloudFormationBaseActor):

    """Deletes a CloudFormation stack

    **Options**

    :name:
      The name of the queue to create

    :region:
      AWS region (or zone) string, like 'us-west-2'

    **Examples**

    .. code-block:: json

       { "desc": "Create production backend stack",
         "actor": "aws.cloudformation.Create",
         "options" {
           "region": "us-west-1",
           "name": "%CF_NAME%",
         }
       }

    **Dry Mode**

    Validates that the CF stack exists, but does not delete it.
    """

    all_options = {
        'name': (str, REQUIRED, 'Name of the stack'),
        'region': (str, REQUIRED, 'AWS region (or zone) name, like us-west-2')
    }

    @gen.coroutine
    def _delete_stack(self):
        """Executes the stack deletion."""
        # Create the stack, and get its ID.
        self.log.info('Deleting stack %s' % self.option('name'))
        try:
            ret = yield self.thread(
                self.cf_conn.delete_stack, self.option('name'))
        except BotoServerError as e:
            msg = '%s: %s' % (e.error_code, e.message)

            if e.status == 400:
                raise CloudFormationError(msg)

            raise
        self.log.info('Stack %s delete requested: %s' %
                      (self.option('name'), ret))
        raise gen.Return(ret)

    @gen.coroutine
    def _execute(self):
        stack_name = self.option('name')

        # If the stack doesn't exist, let the user know.
        exists = yield self._get_stack(stack_name)
        if not exists:
            raise StackNotFound('Stack %s does not exist!' % stack_name)

        # If we're in dry mode, exit at this point. We can't do anything
        # further to validate that the creation process will work.
        if self._dry:
            self.log.info('Skipping CloudFormation Stack deletion.')
            raise gen.Return()

        # Delete
        yield self._delete_stack()

        # Now wait until the stack creation has finished
        try:
            yield self._wait_until_state(DELETED)
        except StackNotFound:
            # Pass here because a stack not found exception is totally
            # reasonable since we're deleting the stack. Sometimes Amazon
            # actually deletes the stack immediately, and othertimes it lists
            # the stack as a 'deleted' state, but we still get that state back.
            # Either case is fine.
            pass

        raise gen.Return()



# In theory, RightScale's API is discoverable through ``links`` in responses.
# In practice, we have to help our robots along with the following hints:
RS_DEFAULT_ACTIONS = {
        'index': {
            'http_method': 'get',
            },
        'show': {
            'http_method': 'get',
            'extra_path': '/%(res_id)s',
            },
        'create': {
            'http_method': 'post',
            },
        'update': {
            'http_method': 'put',
            'extra_path': '/%(res_id)s',
            },
        'destroy': {
            'http_method': 'delete',
            'extra_path': '/%(res_id)s',
            },
        }

ALERT_ACTIONS = {
        'disable': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/disable',
            },
        'enable': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/enable',
            },
        'quench': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/quench',
            },
        'create': None,
        'update': None,
        'destroy': None,
        }

COOKBOOK_ATTACHMENT_ACTIONS = {
        'multi_attach': {
            'http_method': 'post',
            'extra_path': '/multi_attach',
            },
        'multi_detach': {
            'http_method': 'post',
            'extra_path': '/multi_detach',
            },
        'update': None,
        }

INPUT_ACTIONS = {
        'multi_update': {
            'http_method': 'put',
            'extra_path': '/multi_update',
            },
        'show': None,
        'create': None,
        'update': None,
        'destroy': None,
        }

INSTANCE_ACTIONS = {
        'launch': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/launch',
            },
        'lock': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/lock',
            },
        'multi_run_executable': {
            'http_method': 'post',
            'extra_path': '/multi_run_executable',
            },
        'multi_terminate': {
            'http_method': 'post',
            'extra_path': '/multi_terminate',
            },
        'reboot': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/reboot',
            },
        'run_executable': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/run_executable',
            },
        'set_custom_lodgement': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/set_custom_lodgement',
            },
        'start': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/start',
            },
        'stop': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/stop',
            },
        'terminate': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/terminate',
            },
        'unlock': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/unlock',
            },
        'create': None,
        'destroy': None,
        }

MULTI_CLOUD_IMAGE_ACTIONS = {
        'clone': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/clone',
            },
        'commit': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/commit',
            },
        }

SERVER_ARRAY_ACTIONS = {
        'clone': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/clone',
            },
        'current_instances': {
            'http_method': 'get',
            'extra_path': '/%(res_id)s/current_instances',
            },
        'launch': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/launch',
            },
        'multi_run_executable': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/multi_run_executable',
            },
        'multi_terminate': {
            'http_method': 'post',
            'extra_path': '/%(res_id)s/multi_terminate',
            },
        }

UPDATE_NONE_ACTIONS = {
        'update': None,
        }

# Specify variations from the default actions defined in RS_DEFAULT_ACTIONS.
# These specs come from http://reference.rightscale.com/api1.5/index.html
ROOT_COLLECTIONS = {

        'account_groups': {
            'create': None,
            'update': None,
            'destroy': None,
            },

        'accounts': {
            'index': None,
            'create': None,
            'update': None,
            'destroy': None,
            },

        # alert_specs use defaults

        'alerts': ALERT_ACTIONS,

        'audit_entries': {
            'append': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/append',
                },
            'detail': {
                'http_method': 'get',
                'extra_path': '/%(res_id)s/detail',
                },
            'destroy': None,
            },

        'backups': {
            'cleanup': {
                'http_method': 'post',
                'extra_path': '/cleanup',
                },
            },

        'child_accounts': {
            'show': None,
            'destroy': None,
            },

        'cloud_accounts': {
            'update': None,
            },

        'clouds': {
            'create': None,
            'update': None,
            'destroy': None,
            },

        # these are only in the 1.5 docs and are not available as hrefs.
        'cookbook_attachments': COOKBOOK_ATTACHMENT_ACTIONS,

        'cookbooks': {
            'follow': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/follow',
                },
            'freeze': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/freeze',
                },
            'obsolete': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/obsolete',
                },
            'create': None,
            'update': None,
            },

        # credentials use defaults

        'deployments': {
            'clone': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/clone',
                },
            'lock': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/lock',
                },
            'servers': {
                'http_method': 'get',
                'extra_path': '/%(res_id)s/servers',
                },
            'unlock': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/unlock',
                },
            },

        'identity_providers': {
            'create': None,
            'update': None,
            'destroy': None,
            },

        'multi_cloud_images': MULTI_CLOUD_IMAGE_ACTIONS,

        # network_gateways use defaults
        # network_option_group_attachments use defaults
        # network_option_groups use defaults
        # networks use defaults

        # oauth2 is a special case just used during auth

        'permissions': {
            'update': None,
            },

        # only in 1.5 api docs, not discoverable via href
        'placement_groups': {
            'update': None,
            },

        'preferences': {
            'create': None,
            },

        'publication_lineages': {
            'index': None,
            'create': None,
            'update': None,
            'destroy': None,
            },

        'publications': {
            'import': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/import',
                },
            'create': None,
            'update': None,
            'destroy': None,
            },

        'repositories': {
            'cookbook_import': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/cookbook_import',
                },
            'refetch': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/refetch',
                },
            'resolve': {
                'http_method': 'post',
                'extra_path': '/resolve',
                },
            },

        'right_scripts': {
            'create': None,
            'update': None,
            'destroy': None,
            },

        # route_tables uses defaults
        # routes uses defaults
        # security_group_rules uses defaults

        # rs api 1.5 returns a link where rel=self for the ``/api/sessions``
        # resource.  sadly, the href=/api/session.  regardless, we don't need
        # it as an attribute because it's where we started off.
        'self': None,

        'server_arrays': SERVER_ARRAY_ACTIONS,

        'server_template_multi_cloud_images': {
            'make_default': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/make_default',
                },
            'update': None,
            },

        'server_templates': {
            'clone': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/clone',
                },
            'commit': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/commit',
                },
            'detect_changes_in_head': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/detect_changes_in_head',
                },
            'publish': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/publish',
                },
            'resolve': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/resolve',
                },
            'swap_repository': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/swap_repository',
                },
            },

        'servers': {
            'clone': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/clone',
                },
            'launch': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/launch',
                },
            'terminate': {
                'http_method': 'post',
                'extra_path': '/%(res_id)s/terminate',
                },
            },

        # workaround inconsistency in rs hateoas
        'sessions': {
            'accounts': {
                'http_method': 'get',
                'extra_path': '/accounts',
                },
            'index': None,
            'show': None,
            'create': None,
            'update': None,
            'destroy': None,
            },

        'tags': {
            'by_resource': {
                'http_method': 'post',
                'extra_path': '/by_resource',
                },
            'by_tag': {
                'http_method': 'post',
                'extra_path': '/by_tag',
                },
            'multi_add': {
                'http_method': 'post',
                'extra_path': '/multi_add',
                },
            'multi_delete': {
                'http_method': 'post',
                'extra_path': '/multi_delete',
                },
            'index': None,
            'show': None,
            'create': None,
            'update': None,
            'destroy': None,
            },

        'users': {
            'destroy': None,
            },

        }

CLOUD_COLLECTIONS = {

        'datacenters': {
            'create': None,
            'update': None,
            'destroy': None,
            },

        'images': {
            'create': None,
            'update': None,
            'destroy': None,
            },

        'instance_types': {
            'create': None,
            'update': None,
            'destroy': None,
            },

        'instances': INSTANCE_ACTIONS,

        'ip_address_bindings': UPDATE_NONE_ACTIONS,

        # ip_addresses uses defaults

        'recurring_volume_attachments': UPDATE_NONE_ACTIONS,

        'security_groups': {
            'update': None,
            },

        'ssh_keys': {
            'update': None,
            },

        # subnets uses defaults

        'volume_attachments': UPDATE_NONE_ACTIONS,

        'volume_snapshots': UPDATE_NONE_ACTIONS,

        }

INSTANCE_COLLECTIONS = {

        'alerts': ALERT_ACTIONS,
        'inputs': INPUT_ACTIONS,

        # instance_custom_lodgements uses defaults

        'monitoring_metrics': {
            'data': {
                'http_method': 'get',
                'extra_path': '/%(res_id)s/data',
                },
            'create': None,
            'update': None,
            'destroy': None,
            },

        # subnets uses defaults

        # TODO: investigate to see how useful tasks is by itself.  i.e. there's
        # no way to index() all tasks for an instance.  regardless, this
        # definition is here at least for completeness.
        'tasks': {
            'show': {
                'http_method': 'get',
                'extra_path': '/live/tasks/%(res_id)s',
                },
            'index': None,
            'create': None,
            'update': None,
            'destroy': None,
            },

        'volume_attachments': UPDATE_NONE_ACTIONS,

        'volumes': {
            'update': None,
            },

        'volume_types': {
            'create': None,
            'update': None,
            'destroy': None,
            },

        }

COOKBOOK_COLLECTIONS = {
        'cookbook_attachments': COOKBOOK_ATTACHMENT_ACTIONS,
        }

DEPLOYMENT_COLLECTIONS = {
        'alerts': ALERT_ACTIONS,
        'inputs': INPUT_ACTIONS,
        'server_arrays': SERVER_ARRAY_ACTIONS,
        }

IP_ADDRESS_COLLECTIONS = {
        'ip_address_bindings': UPDATE_NONE_ACTIONS,
        }

REPOSITORY_COLLECTIONS = {
        'repository_assets': {
            'create': None,
            'update': None,
            'destroy': None,
            },
        }

SERVER_COLLECTIONS = {
        # alert_specs use defaults
        'alerts': ALERT_ACTIONS,
        }

SERVER_ARRAY_COLLECTIONS = {
        # alert_specs use defaults
        'alerts': ALERT_ACTIONS,
        'current_instances': INSTANCE_ACTIONS,
        }

SERVER_TEMPLATES_COLLECTIONS = {

        'cookbook_attachments': COOKBOOK_ATTACHMENT_ACTIONS,
        'inputs': INPUT_ACTIONS,
        'multi_cloud_images': MULTI_CLOUD_IMAGE_ACTIONS,

        'runnable_bindings': {
            'multi_update': {
                'http_method': 'put',
                'extra_path': '/multi_update',
                },
            'update': None,
            },

        }

VOLUME_COLLECTIONS = {
        'recurring_volume_attachments': UPDATE_NONE_ACTIONS,
        'volume_snapshots': UPDATE_NONE_ACTIONS,
        }

VOLUME_SNAPSHOT_COLLECTIONS = {
        'recurring_volume_attachments': UPDATE_NONE_ACTIONS,
        }

COLLECTIONS = {
        'application/vnd.rightscale.session+json': ROOT_COLLECTIONS,
        'application/vnd.rightscale.cookbook+json': COOKBOOK_COLLECTIONS,
        'application/vnd.rightscale.cloud+json': CLOUD_COLLECTIONS,
        'application/vnd.rightscale.instance+json': INSTANCE_COLLECTIONS,
        'application/vnd.rightscale.ip_address+json': IP_ADDRESS_COLLECTIONS,
        'application/vnd.rightscale.deployment+json': DEPLOYMENT_COLLECTIONS,

        # multi_cloud_image has a ``settings`` collection (a.k.a.
        # MultiCloudImageSettings in the RS docs) that just uses defaults, so
        # no need for an extra map

        'application/vnd.rightscale.repository+json': REPOSITORY_COLLECTIONS,
        'application/vnd.rightscale.server+json': SERVER_COLLECTIONS,
        'application/vnd.rightscale.server_array+json': SERVER_ARRAY_COLLECTIONS,
        'application/vnd.rightscale.server_template+json': SERVER_TEMPLATES_COLLECTIONS,

        # security_group has a ``security_group_rules`` collection that just
        # uses defaults, so no need for an extra map

        'application/vnd.rightscale.volume+json': VOLUME_COLLECTIONS,
        'application/vnd.rightscale.volume_snapshot+json': VOLUME_SNAPSHOT_COLLECTIONS,
        }



from functools import partial
import logging
import time
import requests


log = logging.getLogger(__name__)
DEFAULT_ROOT_RES_PATH = '/'


class HTTPResponse(object):
    """
    Wrapper around :class:`requests.Response`.

    Parses ``Content-Type`` header and makes it available as a list of fields
    in the :attr:`content_type` member.
    """
    def __init__(self, raw_response):
        self.raw_response = raw_response

        content_type = raw_response.headers.get('content-type', '')
        ct_fields = [f.strip() for f in content_type.split(';')]
        self.content_type = ct_fields

    def __getattr__(self, name):
        return getattr(self.raw_response, name)


class HTTPClient(object):

    """
    Convenience wrapper around Requests.

    :param str endpoint: URL for the API endpoint. E.g. ``https://blah.org``.

    :param dict extra_headers: When specified, these key-value pairs are added
        to the default HTTP headers passed in with each request.

    """

    def __init__(
            self,
            endpoint='',
            extra_headers=None,
            oauth_path=None,
            refresh_token=None,
            ):
        self.endpoint = endpoint

        s = requests.session()

        # Disable keepalives. They're unsafe in threaded apps that potentially
        # re-use very old connection objects from the urllib3 connection pool.
        s.headers['Accept'] = 'application/json'
        s.headers['Connection'] = 'close'
        if extra_headers:
            s.headers.update(extra_headers)
        self.s = s

        # convenience methods
        self.delete = partial(self.request, 'delete')
        self.get = partial(self.request, 'get')
        self.head = partial(self.request, 'head')
        self.post = partial(self.request, 'post')
        self.put = partial(self.request, 'put')

        # keep track of when our auth token expires
        self.oauth_path = oauth_path
        self.refresh_token = refresh_token
        self.auth_expires_at = None

    def login(self):
        """
        Gets and stores an OAUTH token from Rightscale.
        """
        log.debug('Logging into RightScale...')
        login_data = {
            'grant_type': 'refresh_token',
            'refresh_token': self.refresh_token,
            }
        response = self._request('post', self.oauth_path, data=login_data)

        raw_token = response.json()
        auth_token = "Bearer %s" % raw_token['access_token']
        self.s.headers['Authorization'] = auth_token

        # Generate an expiration time for our token of 60-seconds before the
        # standard time returned by RightScale. This will be used in the
        # self.client property to validate that our token is still usable on
        # every API call.
        log.debug('Auth Token expires in %s(s)' % raw_token['expires_in'])
        self.auth_expires_at = time.time() + int(raw_token['expires_in']) - 60

    def request(self, method, path='/', url=None, ignore_codes=[], **kwargs):
        """
        Wrapper for the ._request method that verifies if we're logged into
        RightScale before making a call, and sanity checks the oauth expiration
        time.

        :param str method: An HTTP method (e.g. 'get', 'post', 'PUT', etc...)

        :param str path: A path component of the target URL.  This will be
            appended to the value of ``self.endpoint``.  If both :attr:`path`
            and :attr:`url` are specified, the value in :attr:`url` is used and
            the :attr:`path` is ignored.

        :param str url: The target URL (e.g.  ``http://server.tld/somepath/``).
            If both :attr:`path` and :attr:`url` are specified, the value in
            :attr:`url` is used and the :attr:`path` is ignored.

        :param ignore_codes: List of HTTP error codes (e.g.  404, 500) that
            should be ignored.  If an HTTP error occurs and it is *not* in
            :attr:`ignore_codes`, then an exception is raised.
        :type ignore_codes: list of int

        :param kwargs: Any other kwargs to pass to :meth:`requests.request()`.

        Returns a :class:`requests.Response` object.
        """
        # On every call, check if we're both logged in, and if the token is
        # expiring. If it is, we'll re-login with the information passed into
        # us at instantiation.
        if time.time() > self.auth_expires_at:
            self.login()

        # Now make the actual API call
        return self._request(method, path, url, ignore_codes, **kwargs)

    def _request(self, method, path='/', url=None, ignore_codes=[], **kwargs):
        """
        Performs HTTP request.

        :param str method: An HTTP method (e.g. 'get', 'post', 'PUT', etc...)

        :param str path: A path component of the target URL.  This will be
            appended to the value of ``self.endpoint``.  If both :attr:`path`
            and :attr:`url` are specified, the value in :attr:`url` is used and
            the :attr:`path` is ignored.

        :param str url: The target URL (e.g.  ``http://server.tld/somepath/``).
            If both :attr:`path` and :attr:`url` are specified, the value in
            :attr:`url` is used and the :attr:`path` is ignored.

        :param ignore_codes: List of HTTP error codes (e.g.  404, 500) that
            should be ignored.  If an HTTP error occurs and it is *not* in
            :attr:`ignore_codes`, then an exception is raised.
        :type ignore_codes: list of int

        :param kwargs: Any other kwargs to pass to :meth:`requests.request()`.

        Returns a :class:`requests.Response` object.
        """
        _url = url if url else (self.endpoint + path)
        r = self.s.request(method, _url, **kwargs)
        if not r.ok and r.status_code not in ignore_codes:
            r.raise_for_status()
        return HTTPResponse(r)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2014 Nextdoor.com, Inc
"""
This package provides a quick way of creating custom API clients for JSON-based
REST APIs. The majority of the work is in the creation of a _CONFIG dictionary
for the class. This dictionary dynamically configures the object at
instantiation time with the appropriate @gen.coroutine wrapped HTTP fetch
methods.

See the documentation in docs/DEVELOPMENT.md for more details on how to use
this package to create your own API client.
"""

import logging
import types
import urllib

from tornado import gen
from tornado import httpclient
from tornado import httputil
import simplejson as json

from kingpin import utils
from kingpin.actors import exceptions

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


def _retry(*f_or_args, **options):
    """Coroutine-compatible Retry Decorator.

    This decorator provides a simple retry mechanism that compares the
    exceptions it received against a configuration list (self._EXCEPTIONS), and
    then performs the action defined in that list. For example, an HTTPError
    with a '500' code might want to retry 3 times. On the otherhand, a 401/403
    might want to throw an InvalidCredentials exception.

    Examples:

    >>> @_retry
        def some_func(self):
            yield ...

    >>> @_retry(retries=5):
        def some_func(self):
            yield ...

    """

    # Defaults...
    retries = 3
    delay = 0.25

    # Have to determine if invoked as @_retry or @_retry()
    if len(f_or_args) == 1 and callable(f_or_args[0]):
        # Decorator invoked as @_retry
        _call_with_args = False
    else:
        # Decorator invoked as @_retry(args...)
        # `f` is unknown for now
        _call_with_args = True
        retries = options.pop('retries', retries)
        delay = options.pop('delay', delay)

    def decorator(f_or_self, *args, **kwargs):
        # Depending on how this decorator is invoked
        # The first argument is either the function, or the `self` object
        if not _call_with_args:
            self = f_or_self
            f = f_or_args[0]
        else:
            f = f_or_self

        def wrapper(self, *args, **kwargs):
            i = 1

            # Get a list of private kwargs to mask
            private_kwargs = getattr(self, '_private_kwargs', [])

            # For security purposes, create a patched kwargs string that
            # removes passwords from the arguments. This is never guaranteed to
            # work (an API could have 'foo' as their password field, and we
            # just won't know ...), but we make a best effort here.
            safe_kwargs = dict(kwargs)
            remove = [k for k in safe_kwargs if k in private_kwargs]
            for k in remove:
                safe_kwargs[k] = '****'

            while True:
                # Don't log out the first try as a 'Try' ... just do it
                if i > 1:
                    log.debug('Try (%s/%s) of %s(%s, %s)' %
                              (i, retries, f, args, safe_kwargs))

                # Attempt the method. Catch any exception listed in
                # self._EXCEPTIONS.

                try:
                    ret = yield gen.coroutine(f)(self, *args, **kwargs)
                    raise gen.Return(ret)
                except tuple(self._EXCEPTIONS.keys()) as e:
                    error = str(e)
                    if hasattr(e, 'message'):
                        error = e.message
                    log.warning('Exception raised on try %s: %s' % (i, error))

                    # If we've run out of retry attempts, raise the exception
                    if i >= retries:
                        log.debug('Raising exception: %s' % e)
                        raise e

                    # Gather the config for this exception-type from
                    # self._EXCEPTIONS. Iterate through the data and see if we
                    # have a matching exception string.
                    exc_conf = self._EXCEPTIONS[type(e)].copy()

                    # An empty string for the key is the default exception
                    # It's optional, but can match before others match, so we
                    # pop it before searching.
                    default_exc = exc_conf.pop('', False)
                    log.debug('Searching through %s' % exc_conf)
                    matched_exc = [exc for key, exc in exc_conf.items()
                                   if key in str(e)]

                    log.debug('Matched exceptions: %s' % matched_exc)
                    if matched_exc and matched_exc[0] is not None:
                        exception = matched_exc[0]
                        log.debug('Matched exception: %s' % exception)
                        raise exception(error)
                    elif matched_exc and matched_exc[0] is None:
                        log.debug('Exception is retryable!')
                        pass
                    elif default_exc is not False:
                        raise default_exc(str(e))
                    elif default_exc is False:
                        # Reaching this part means no exception was matched
                        # and no default was specified.
                        log.debug('No explicit behavior for this exception'
                                  ' found. Raising.')
                        raise e

                    # Must have been a retryable exception. Retry.
                    i = i + 1
                    log.debug('Retrying in %s...' % delay)
                    yield utils.tornado_sleep(delay)

                log.debug('Retrying..')

        if not _call_with_args:
            # Invoked as @_retry
            # Should return the evaluated run
            return wrapper(self, *args, **kwargs)
        else:
            # Invoked as @_retry(args...)
            # Return a wrapper that expects all the new args
            return wrapper

    return decorator


def create_http_method(name, http_method):
    """Creates the get/put/delete/post coroutined-method for a resource.

    This method is called during the __init__ of a RestConsumer object. The
    method creates a custom method thats handles a GET, PUT, POST or DELETE
    through the Tornado HTTPClient class.

    Args:
        http_method: Name of the method (get, put, post, delete)

    Returns:
        A method appropriately configured and named.
    """

    @gen.coroutine
    def method(self, *args, **kwargs):
        # We don't support un-named args. Throw an exception.
        if args:
            raise exceptions.InvalidOptions('Must pass named-args (kwargs)')

        ret = yield self._client.fetch(
            url='%s%s' % (self._ENDPOINT, self._path),
            method=http_method.upper(),
            params=kwargs,
            auth_username=self._CONFIG.get('auth', {}).get('user'),
            auth_password=self._CONFIG.get('auth', {}).get('pass')
        )
        raise gen.Return(ret)

    method.__name__ = http_method
    return method


def create_method(name, config):
    """Creates a RestConsumer object.

    Configures a fresh RestConsumer object with the supplied configuration
    bits. The configuration includes information about the name of the method
    being consumed and the configuration of that method (which HTTP methods it
    supports, etc).

    The final created method accepts any args (`*args, **kwargs`) and passes
    them on to the RestConsumer object being created. This allows for passing
    in unique resource identifiers (ie, the '%res%' in
    '/v2/rooms/%res%/history').

    Args:
        name: The name passed into the RestConsumer object
        config: The config passed into the RestConsumer object

    Returns:
        A method that returns a fresh RestConsumer object
    """

    def method(self, *args, **kwargs):
        # Merge the supplied kwargs to the method with any kwargs supplied to
        # the RestConsumer parent object. This ensures that tokens replaced in
        # the 'path' variables are passed all the way down the instantiation
        # chain.
        merged_kwargs = dict(self._kwargs.items() + kwargs.items())

        return self.__class__(
            name=name,
            config=self._attrs[name],
            client=self._client,
            *args, **merged_kwargs)

    method.__name__ = name
    return method


class RestConsumer(object):

    """An abstract object that self-defines its own API access methods.

    At init time, this object reads its `_CONFIG` and pre-defines all of the
    API access methods that have been described. It does not handle actual HTTP
    calls directly, but is passed in a `client` object (anything that
    subclasses the RestClient class) and leverages that for the actual web
    calls.
    """

    _CONFIG = {}
    _ENDPOINT = None

    def __init__(self, name=None, config=None, client=None, *args, **kwargs):
        """Initialize the RestConsumer object.

        The generic RestConsumer object (with no parameters passed in) looks at
        the self.__class__._CONFIG dictionary and dynamically generates access
        methods for the various API methods.

        The GET, PUT, POST and DELETE methods optionally listed in
        CONFIG['http_methods'] represent the possible types of HTTP methods
        that the CONFIG['path'] supports. For each one of these listed, a
        @coroutine wrapped get/put/post/delete() method will be created in the
        RestConsumer that knows how to make the HTTP request.

        For each item listed in CONFIG['attrs'], an access method is created
        that will dynamically create and return a new RestConsumer object thats
        configured for this endpoint. These methods are not asynchronous, but
        are non-blocking.

        Args:
            name: Name of the resource method (default: None)
            config: The dictionary object with the configuration for this API
                    endpoint call.
            client: <TBD>
            *args,**kwargs: <TBD>
        """
        # If these aren't passed in, then get them from the class definition
        name = name or self.__class__.__name__
        config = config or self._CONFIG

        # Get the basic options for this particular REST endpoint access object
        self._path = config.get('path', None)
        self._http_methods = config.get('http_methods', None)
        self._attrs = config.get('attrs', None)
        self._kwargs = kwargs

        # If no client was supplied, then we
        self._client = client or RestClient()

        # Ensure that any tokens that need filling-in in the self._path setting
        # are pulled from the **kwargs passed into this init. This is used on
        # API paths like Hipchats '/v2/room/%(res)/...' URLs.
        self._path = self._replace_path_tokens(self._path, kwargs)

        # Create all of the methods based on the self._http_methods and
        # self._attrs dict.
        self._create_methods()
        self._create_attrs()

        # Log some things
        log.debug('%s/%s initialized' %
                  (self.__class__.__name__, self._client))

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self)

    def __str__(self):
        return str(self._path)

    def _replace_path_tokens(self, path, tokens):
        """Search and replace %xxx% with values from tokens.

        Used to replace any values of %xxx% with 'xxx' from tokens. Can replace
        one, or many fields at aonce.

        Args:
            path: String of the path
            tokens: A dictionary of tokens to search through.

        Returns:
            path: A modified string
        """
        if not path:
            return

        try:
            path = utils.populate_with_tokens(path, tokens)
        except LookupError as e:
            msg = 'Path (%s), tokens: (%s) error: %s' % (path, tokens, e)
            raise TypeError(msg)

        return path

    def _create_methods(self):
        """Create @gen.coroutine wrapped HTTP methods.

        Iterates through the methods described in self._methods and creates
        @gen.coroutine wrapped access methods that perform these actions.
        """
        if not self._http_methods:
            return

        for name in self._http_methods.keys():
            full_method_name = 'http_%s' % name
            method = create_http_method(full_method_name, name)
            setattr(self,
                    full_method_name,
                    types.MethodType(method, self, self.__class__))

    def _create_attrs(self):
        """Creates access methods to the attributes in self._attrs.

        Iterates through the attributes described in self._attrs and creates
        access methods that return RestConsumer objects for those attributes.
        """
        if not self._attrs:
            return

        for name in self._attrs.keys():
            method = create_method(name, self._attrs[name])
            setattr(self, name, types.MethodType(method, self, self.__class__))


class RestClient(object):

    """Very simple REST client for the RestConsumer. Implements a
    AsyncHTTPClient(), some convinience methods for URL escaping, and a single
    fetch() method that can handle GET/POST/PUT/DELETEs.

    This code is nearly identical to the kingpin.actors.base.BaseHTTPActor
    class, but is not actor-specific.

    Args:
        headers: Headers to pass in on every HTTP request
    """

    _EXCEPTIONS = {
        httpclient.HTTPError: {
            '401': exceptions.InvalidCredentials,
            '403': exceptions.InvalidCredentials,
            '500': None,
            '502': None,
            '503': None,
            '504': None,

            # Rrepresents a standard HTTP Timeout
            '599': None,

            '': exceptions.RecoverableActorFailure,
        }
    }

    def __init__(self, client=None, headers=None):
        self._client = client or httpclient.AsyncHTTPClient()
        self._private_kwargs = ['auth_password']
        self.headers = headers

    def _generate_escaped_url(self, url, args):
        """Takes in a dictionary of arguments and returns a URL line.

        Sorts the arguments so that the returned string is predictable and in
        alphabetical order. Effectively wraps the tornado.httputil.url_concat
        method and properly strips out None values, as well as lowercases
        Bool values.

        Args:
            url: (Str) The URL to append the arguments to
            args: (Dict) Key/Value arguments. Values should be primitives.

        Returns:
            A URL encoded string like this: <url>?foo=bar&abc=xyz
        """

        # Remove keys from the arguments where the value is None
        args = dict((k, v) for k, v in args.iteritems() if v)

        # Convert all Bool values to lowercase strings
        for key, value in args.iteritems():
            if type(value) is bool:
                args[key] = str(value).lower()

        # Now generate the URL
        full_url = httputil.url_concat(url, sorted(args.items()))
        log.debug('Generated URL: %s' % full_url)

        return full_url

    # TODO: Add a retry/backoff timer here. If the remote endpoint returns
    # garbled data (ie, maybe a 500 errror or something else thats not in
    # JSON format, we should back off and try again.
    @gen.coroutine
    @_retry
    def fetch(self, url, method, params={},
              auth_username=None, auth_password=None):
        """Executes a web request asynchronously and yields the body.

        Args:
            url: (Str) The full url path of the API call
            params: (Dict) Arguments (k/v pairs) to submit either as POST data
                    or URL argument options.
            method: (Str) GET/PUT/POST/DELETE
            auth_username: (str) HTTP auth username
            auth_password: (str) HTTP auth password
        """

        # Start with empty post data. If we're doing a PUT/POST, then just pass
        # args directly into the ch() method and let it take care of
        # things. If we're doing a GET/DELETE though, convert kwargs into a
        # modified URL string and pass that into the fetch() method.
        body = None
        if method in ('PUT', 'POST'):
            body = urllib.urlencode(params) or None
        elif method in ('GET', 'DELETE') and params:
            url = self._generate_escaped_url(url, params)

        # Generate the full request URL and log out what we're doing...
        log.debug('Making %s request to %s. Data: %s' % (method, url, body))

        # Create the http_request object
        http_request = httpclient.HTTPRequest(
            url=url,
            method=method,
            body=body,
            headers=self.headers,
            auth_username=auth_username,
            auth_password=auth_password,
            follow_redirects=True,
            max_redirects=10)

        # Execute the request and raise any exception. Exceptions are not
        # caught here because they are unique to the API endpoints, and thus
        # should be handled by the individual Actor that called this method.
        log.debug('HTTP Request: %s' % http_request)
        try:
            http_response = yield self._client.fetch(http_request)
        except httpclient.HTTPError as e:
            log.critical('Request for %s failed: %s' % (url, e))
            raise
        log.debug('HTTP Response: %s' % http_response.body)

        try:
            body = json.loads(http_response.body)
        except ValueError:
            raise gen.Return(http_response.body)

        # Receive a successful return
        raise gen.Return(body)


class SimpleTokenRestClient(RestClient):

    """Simple RestClient that appends a 'token' to every web request for
    authentication. Used in most simple APIs where a token is provided to the
    end user.

    Args:
        tokens: (dict) A dict with the token name/value(s) to append to every
                we request.
    """

    def __init__(self, tokens, *args, **kwargs):
        super(SimpleTokenRestClient, self).__init__(*args, **kwargs)
        self._tokens = tokens
        for key in tokens.keys():
            self._private_kwargs.append(key)

    @gen.coroutine
    def fetch(self, *args, **kwargs):
        if 'params' not in kwargs:
            kwargs['params'] = {}

        kwargs['params'].update(self._tokens)
        ret = yield super(SimpleTokenRestClient, self).fetch(*args, **kwargs)
        raise gen.Return(ret)



"""
Python-Rightscale

A stupid wrapper around rightscale's HTTP API
"""
import types
from .actions import RS_DEFAULT_ACTIONS, COLLECTIONS
from .httpclient import HTTPClient
from .util import get_rc_creds, HookList


# magic strings from the 1.5 api
DEFAULT_API_PREPATH = '/api'

# authenticate here
OAUTH2_RES_PATH = '/'.join((DEFAULT_API_PREPATH, 'oauth2'))

# start hypermedia searches here
ROOT_RES_PATH = '/'.join((DEFAULT_API_PREPATH, 'sessions'))

# these *should* be discoverable from the '/api/sessions' route above, but they
# are not.  there is an open ticket filed with rightscale.  until it gets
# addressed, it's just more magic:
ACCOUNT_INFO_RES_PATH = '/'.join((DEFAULT_API_PREPATH, 'sessions/accounts'))
HEALTH_CHECK_RES_PATH = '/'.join((DEFAULT_API_PREPATH, 'health-check'))

COLLECTION_TYPE = 'type=collection'


def get_resource_method(name, template):
    """
    Creates a function that is suitable as a method for ResourceCollection.
    """
    def rsr_meth(self, **kwargs):
        http_method = template['http_method']
        extra_path = template.get('extra_path')
        if extra_path:
            fills = {'res_id': kwargs.pop('res_id', '')}
            path = self.path + (extra_path % fills)
        else:
            path = self.path
        response = self.client.request(http_method, path, **kwargs)

        loc = response.headers.get('location', None)

        if loc:
            # If the returned code is a 201, then there should be a location
            # header in the response that we can use to re-get the newly created
            # resource.
            loc = response.headers.get('location')
            response = self.client.get(loc, **kwargs)

        # At this point, we better have a valid JSON response object
        try:
            obj = response.json()
        except:
            # The response had no JSON ... not a resource object
            return

        if COLLECTION_TYPE in response.content_type:
            ret = HookList(
                    [Resource(r, path, response, self.client) for r in obj],
                    response=response
                    )
        else:
            ret = Resource(obj, path, response, self.client)
        return ret

    rsr_meth.__name__ = name
    return rsr_meth


class Resource(object):
    """
    A single resource.

    :param dict soul: The essence of the resource as returned by the RightScale
        API.  This is the dictionary of attributes originally returned as the
        JSON body of the HTTP response from RightScale.

    :param str path: The path portion of the URL.  E.g. ``/api/clouds/1``.

    :param rightscale.httpclient.HTTPResponse response: The raw response object
        returned by :meth:`HTTPClient.request`.

    """
    def __init__(self, soul=None, path='', response=None, client=None):
        if soul is None:
            soul = {}
        self.soul = soul
        self.path = path
        self.collection_actions = {}
        self.response = response
        self.client = client
        self._links = None

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.soul)

    def __str__(self):
        return str(self.soul)

    def __cmp__(self, other):
        return cmp(self.soul, other.soul)

    @property
    def content_type(self):
        if self.response:
            return self.response.content_type[0]
        return ''

    def _get_rel_hrefs(self):
        rel_hrefs = self.soul.get('links', [])
        return dict((raw['rel'], raw['href']) for raw in rel_hrefs)

    @property
    def href(self):
        return self._get_rel_hrefs().get('self', '')

    @property
    def links(self):
        # only initialize once, not if empty
        if self._links is None:
            _links = self._get_rel_hrefs()

            collection_actions = COLLECTIONS.get(self.content_type, {})
            self.collection_actions = collection_actions
            for name, action in collection_actions.iteritems():
                if action is None and name in _links:
                    del _links[name]
                    continue
                if name not in _links:
                    _links[unicode(name)] = unicode(
                            '%s/%s' % (self.path, name)
                            )

            self._links = _links
        return self._links

    def __dir__(self):
        return self.links.keys()

    def __getattr__(self, name):
        path = self.links.get(name)
        if not path:
            raise AttributeError('%s object has no attribute %s' % (
                self.__class__.__name__,
                name,
                ))
        actions = RS_DEFAULT_ACTIONS.copy()
        tpl = self.collection_actions.get(name)
        if tpl:
            actions.update(tpl)
        return ResourceCollection(path, self.client, actions)


class ResourceCollection(object):
    def __init__(self, path, client, actions):
        self.path = path
        self.client = client
        for name, template in actions.items():
            if not template:
                continue
            method = get_resource_method(name, template)
            setattr(self, name, types.MethodType(method, self, self.__class__))


class RightScale(Resource):

    def __init__(
            self,
            path=DEFAULT_API_PREPATH,
            refresh_token=None,
            api_endpoint=None,
            ):
        """
        Creates and configures the API object.
        :param str refresh_token: The refresh token provided by Rightscale when
            API access is enabled.
        :param api_endpoint: The rightscale subdomain to be hit with API
            requests.
        :param str path: The path portion of the URL.
            E.g. ``/api``.
        """
        super(RightScale, self).__init__({}, path)
        self.auth_token = None

        rc_creds = get_rc_creds()

        # prevent dumb leakage from the environment by only grabbing creds from
        # rc file if they are not specified to the constructor.
        if api_endpoint is None:
            api_endpoint = rc_creds[0]
        if not api_endpoint:
            raise ValueError("Can't login with no api endpoint.")
        self.api_endpoint = api_endpoint

        if refresh_token is None:
            refresh_token = rc_creds[1]
        if not refresh_token:
            raise ValueError("Can't login. Need refresh token!")

        self.client = HTTPClient(
                api_endpoint,
                {'X-API-Version': '1.5'},
                OAUTH2_RES_PATH,
                refresh_token,
                )

    def health_check(self):
        # only in 1.5 api docs, not discoverable via href
        return self.client.get(HEALTH_CHECK_RES_PATH).json()

    @property
    def links(self):
        if not self.soul:
            try:
                response = self.client.get(ROOT_RES_PATH)
                self.response = response
                self.soul = response.json()
            except:
                self.soul = {}
        return super(RightScale, self).links



import logging

from tornado import httpclient
from tornado import testing
import mock

from kingpin.actors import exceptions
from kingpin.actors import misc
from kingpin.actors.test.helper import mock_tornado

log = logging.getLogger(__name__)


class TestMacro(testing.AsyncTestCase):

    def setUp(self):
        super(TestMacro, self).setUp()
        reload(misc)

    def test_init(self):
        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_macro = mock.Mock(return_value='unit-test-macro')

        with mock.patch('kingpin.utils.convert_json_to_dict') as j2d, \
                mock.patch('kingpin.schema.validate') as schema_validate, \
                mock.patch('kingpin.actors.utils.get_actor') as get_actor:

            j2d.return_value = {
                'desc': 'unit test',
                'actor': 'unit test',
                'options': {}
            }

            actor = misc.Macro('Unit Test', {'macro': 'test.json',
                                             'tokens': {}})

            j2d.assert_called_with(json_file='unit-test-macro', tokens={})
            self.assertEquals(schema_validate.call_count, 1)
            self.assertEquals(actor.initial_actor, get_actor())

    def test_init_remote(self):
        misc.Macro._get_config_from_json = mock.Mock()
        misc.Macro._check_schema = mock.Mock()
        with mock.patch('kingpin.actors.utils.get_actor'):
            with mock.patch.object(httpclient.HTTPClient, 'fetch'):
                misc.Macro('Unit Test', {'macro': 'http://test.json',
                                         'tokens': {}})

    def test_init_dry(self):
        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_config_from_json = mock.Mock()
        misc.Macro._check_schema = mock.Mock()

        with mock.patch('kingpin.utils.convert_json_to_dict') as j2d, \
                mock.patch('kingpin.schema.validate'), \
                mock.patch('kingpin.actors.utils.get_actor'):
            j2d.return_value = {
                'desc': 'unit test',
                'actor': 'unit test',
                'options': {}
            }

            actor = misc.Macro('Unit Test',
                               {'macro': 'examples/test/sleep.json',
                                'tokens': {}},
                               dry=True)

            self.assertTrue(actor.initial_actor._dry)

    def test_init_with_errors(self):

        # Remote files are prohibited for now
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro('Unit Test', {'macro': 'ftp://fail.test.json',
                                     'tokens': {}})

        # Remote file with bad URL
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro('Unit Test', {'macro': 'http://fail.test.json',
                                     'tokens': {}})

        # Non-existent file
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro('Unit Test', {'macro': 'dontcreatethis.json',
                                     'tokens': {}})

        # We don't want the rest of the tests failing on downloading this file.
        misc.Macro._get_macro = mock.Mock(return_value='unit-test-file')

        # Schema failure
        with mock.patch('kingpin.utils.convert_json_to_dict') as j2d:
            j2d.return_value = {
                'desc': 'unit test',
                'options': {}  # `actor` keyword is missing
            }

            with self.assertRaises(exceptions.UnrecoverableActorFailure):
                misc.Macro('Unit Test', {'macro': 'test.json',
                                         'tokens': {}})

        # JSON syntax error
        with mock.patch('kingpin.utils.convert_json_to_dict') as j2d:

            j2d.side_effect = Exception('Something failed with JSON')

            with self.assertRaises(exceptions.UnrecoverableActorFailure):
                misc.Macro('Unit Test', {'macro': 'test.json',
                                         'tokens': {}})

    @testing.gen_test
    def test_execute(self):

        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_macro = mock.Mock()
        misc.Macro._get_config_from_json = mock.Mock()
        misc.Macro._check_schema = mock.Mock()

        with mock.patch('kingpin.actors.utils.get_actor') as get_actor:
            actor = misc.Macro('Unit Test', {'macro': 'test.json',
                                             'tokens': {}},
                               dry=True)

            get_actor().execute = mock_tornado()
            yield actor._execute()

            self.assertEquals(get_actor().execute._call_count, 1)


class TestSleep(testing.AsyncTestCase):

    @testing.gen_test
    def test_execute(self):
        # Call the executor and test it out
        actor = misc.Sleep('Unit Test Action', {'sleep': 0.1})
        res = yield actor.execute()

        # Make sure we fired off an alert.
        self.assertEquals(res, None)

    @testing.gen_test
    def test_execute_with_str(self):
        # Call the executor and test it out
        actor = misc.Sleep('Unit Test Action', {'sleep': '0.5'})
        res = yield actor.execute()

        # Make sure we fired off an alert.
        self.assertEquals(res, None)


class TestGenericHTTP(testing.AsyncTestCase):

    @testing.gen_test
    def test_execute_dry(self):
        actor = misc.GenericHTTP('Unit Test Action',
                                 {'url': 'http://example.com'},
                                 dry=True)

        actor._fetch = mock_tornado()

        yield actor.execute()

        self.assertEquals(actor._fetch._call_count, 0)

    @testing.gen_test
    def test_execute(self):
        actor = misc.GenericHTTP('Unit Test Action',
                                 {'url': 'http://example.com'})
        actor._fetch = mock_tornado({'success': {'code': 200}})

        yield actor.execute()

    @testing.gen_test
    def test_execute_fail(self):
        actor = misc.GenericHTTP('Unit Test Action',
                                 {'url': 'http://example.com'})
        error = httpclient.HTTPError(code=401, response={})
        actor._fetch = mock_tornado(exc=error)

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()



import logging

from tornado import gen
from tornado import testing

from kingpin.actors import exceptions
from kingpin.actors import utils
from kingpin.actors import misc


log = logging.getLogger(__name__)


class FakeActor(object):

    """Fake Actor use for Unit Tests"""

    def __init__(self, desc, options, dry=False):
        log.info('Initializing %s: %s' % (desc, options))
        self.options = options

    @gen.coroutine
    def execute(self):
        raise gen.Return(self.options['return_value'])


class TestUtils(testing.AsyncTestCase):

    def test_get_actor(self):
        actor_return_true = {
            'desc': 'returns true',
            'actor': 'kingpin.actors.test.test_utils.FakeActor',
            'options': {'return_value': True}}
        ret = utils.get_actor(actor_return_true, dry=True)
        self.assertEquals(True, ret.options['return_value'])
        self.assertEquals(FakeActor, type(ret))

    def test_get_actor_class(self):
        actor_string = 'misc.Sleep'
        ret = utils.get_actor_class(actor_string)
        self.assertEquals(type(misc.Sleep), type(ret))

    def test_get_actor_class_direct(self):
        actor_string = 'kingpin.actors.test.test_utils.FakeActor'
        ret = utils.get_actor_class(actor_string)
        self.assertEquals(type(FakeActor), type(ret))

    def test_get_actor_class_bogus_actor(self):
        actor_string = 'bogus.actor'
        with self.assertRaises(exceptions.InvalidActor):
            utils.get_actor_class(actor_string)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2014 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.base`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The RightScale Actors allow you to interact with resources inside your
Rightscale account. These actors all support dry runs properly, but each
actor has its own caveats with ``dry=True``. Please read the instructions
below for using each actor.

**Required Environment Variables**

:RIGHTSCALE_TOKEN:
  RightScale API Refresh Token
  (from the *Account Settings/API Credentials* page)

:RIGHTSCALE_ENDPOINT:
  Your account-specific API Endpoint
  (defaults to https://my.rightscale.com)
"""

from random import randint
import collections
import logging
import os

from tornado import gen
import mock

from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.actors.rightscale import api

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


TOKEN = os.getenv('RIGHTSCALE_TOKEN', None)
ENDPOINT = os.getenv('RIGHTSCALE_ENDPOINT', 'https://my.rightscale.com')


class ArrayNotFound(exceptions.RecoverableActorFailure):

    """Raised when a ServerArray could not be found."""


class ArrayAlreadyExists(exceptions.RecoverableActorFailure):

    """Raised when a ServerArray already exists by a given name."""


class RightScaleBaseActor(base.BaseActor):

    """Abstract class for creating RightScale cloud actors."""

    def __init__(self, *args, **kwargs):
        """Initializes the Actor."""
        super(RightScaleBaseActor, self).__init__(*args, **kwargs)

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "RIGHTSCALE_TOKEN" environment variable.')

        self._client = api.RightScale(token=TOKEN, endpoint=ENDPOINT)

    @gen.coroutine
    def _find_server_arrays(self, array_name,
                            raise_on='notfound',
                            allow_mock=True,
                            exact=True):
        """Find a ServerArray by name and return it.

        Args:
            array_name: String name of the ServerArray to find.
            raise_on: Either None, 'notfound' or 'found'
            allow_mock: Boolean whether or not to allow a Mock object to be
                        returned instead.
            exact: Boolean whether or not to allow multiple arrays to be
                   returned.

        Raises:
            gen.Return(<rightscale.Resource of Server Array>)
            ArrayNotFound()
            ArrayAlreadyExists()
        """
        if raise_on == 'notfound':
            msg = 'Verifying that array "%s" exists' % array_name
        elif raise_on == 'found':
            msg = 'Verifying that array "%s" does not exist' % array_name
        elif not raise_on:
            msg = 'Searching for array named "%s"' % array_name
        else:
            raise exceptions.UnrecoverableActorFailure(
                'Invalid "raise_on" setting in actor code.')

        self.log.debug(msg)
        array = yield self._client.find_server_arrays(array_name, exact=exact)

        if not array and self._dry and allow_mock:
            # Create a fake ServerArray object thats mocked up to help with
            # execution of the rest of the code.
            self.log.info('Array "%s" not found -- creating a mock.' %
                          array_name)
            array = mock.MagicMock(name=array_name)
            # Give the mock a real identity and give it valid elasticity
            # parameters so the Launch() actor can behave properly.
            array.soul = {
                # Used elsewhere to know whether we're working on a mock
                'fake': True,

                # Fake out common server array object properties
                'name': '<mocked array %s>' % array_name,
                'elasticity_params': {'bounds': {'min_count': 4}}
            }
            array.self.path = '/fake/array/%s' % randint(10000, 20000)
            array.self.show.return_value = array

        if array and raise_on == 'found':
            raise ArrayAlreadyExists('Array "%s" already exists!' % array_name)

        if not array and raise_on == 'notfound':
            raise ArrayNotFound('Array "%s" not found!' % array_name)

        # Quick note. If many arrays were returned, lets make sure we throw a
        # note to the user so they know whats going on.
        if isinstance(array, list):
            for a in array:
                self.log.info('Matching array found: %s' % a.soul['name'])

        raise gen.Return(array)

    def _generate_rightscale_params(self, prefix, params):
        """Utility function for creating RightScale-style parameters.

        RightScale takes inputs in the form of a hash of key/value pairs, but
        these pairs are in a strange pseudo-dict form. This method takes a
        standard hash and converts it into a rightscale-compatible form.

        For example, take this dict:

            {'name': 'unittest-name',
             'bounds': { 'min_count': 3}

        We return:

            [ ('server_array[name]', 'unittest-name'),
              ('server_array[bounds][min_count]', '3) ]

        For more examples, see our unit tests.

        Args:
            prefix: The key-prefix to use (ie, 'server_array')
            params: The dictionary to squash

        Returns:
            A list of tuples of key/value pairs.
        """
        if not type(params) == dict:
            raise exceptions.InvalidOptions(
                'Parameters passed in must be in the form of a dict.')

        # Nested loop that compresses a multi level dictinary into a flat
        # array of key=value strings.
        def flatten(d, parent_key=prefix, sep='_'):
            items = []

            if isinstance(d, collections.MutableMapping):
                # If a dict is passed in, break it into its items and
                # then iterate over them.
                for k, v in d.items():
                    new_key = parent_key + '[' + k + ']' if parent_key else k
                    items.extend(flatten(v, new_key))
            elif isinstance(d, list):
                # If an array was passed in, then iterate over the array
                new_key = parent_key + '[]' if parent_key else k
                for item in d:
                    items.extend(flatten(item, new_key))
            else:
                items.append((parent_key, d))

            return items

        return sorted(flatten(params))



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2014 Nextdoor.com, Inc
"""CLI Script Runner for Kingpin."""

import logging
import optparse
import os
import sys

from tornado import gen
from tornado import ioloop

from kingpin import utils
from kingpin.actors import exceptions as actor_exceptions
from kingpin.actors.misc import Macro
from kingpin.version import __version__


log = logging.getLogger(__name__)

__author__ = 'Matt Wise (matt@nextdoor.com)'


# We handle all the exceptions ourselves, so additional log statements from
# BOTO are not needed.
logging.getLogger('boto').setLevel(logging.CRITICAL)

# Initial option handler to set up the basic application environment.
usage = 'usage: %prog [json file] <options>'
parser = optparse.OptionParser(usage=usage, version=__version__,
                               add_help_option=True)
parser.set_defaults(verbose=True)

# Job Configuration
parser.add_option('-j', '--json', dest='json',
                  help='Path to JSON Deployment File')
parser.add_option('-d', '--dry', dest='dry', action='store_true',
                  help='Executes a dry run only.')

# Logging Configuration
parser.add_option('-l', '--level', dest='level', default='info',
                  help='Set logging level (INFO|WARN|DEBUG|ERROR)')
parser.add_option('', '--debug', dest='level_debug', default=False,
                  action='store_true', help='Equivalent to --level=DEBUG')
parser.add_option('-c', '--color', dest='color', default=False,
                  action='store_true', help='Colorize the log output')

(options, args) = parser.parse_args()


def kingpin_fail(message):
    parser.print_help()
    sys.stderr.write('\nError: %s\n' % message)
    sys.exit(1)


@gen.coroutine
def main():

    env_tokens = dict(os.environ)

    try:
        json_file = options.json or sys.argv[1] if sys.argv else None
    except Exception as e:
        kingpin_fail(
            '%s You must specify --json or provide it as first argument.' % e)

    # Begin doing real stuff!
    if os.environ.get('SKIP_DRY', False):
        log.warn('')
        log.warn('*** You have disabled the dry run.')
        log.warn('*** Execution will begin with no expectation of success.')
        log.warn('')
    elif not options.dry:
        log.info('Rehearsing... Break a leg!')
        try:
            dry_actor = Macro(desc='Kingpin',
                              options={'macro': json_file,
                                       'tokens': env_tokens},
                              dry=True)
            yield dry_actor.execute()
        except actor_exceptions.ActorException as e:
            log.critical('Dry run failed. Reason:')
            log.critical(e)
            sys.exit(2)
        else:
            log.info('Rehearsal OK! Performing!')

    try:
        log.info('Lights, camera ... action!')
        runner = Macro(desc='Kingpin',
                       options={'macro': json_file,
                                'tokens': env_tokens},
                       dry=options.dry)
        yield runner.execute()
    except actor_exceptions.ActorException as e:
        log.error('Kingpin encountered mistakes during the play.')
        log.error(e)
        sys.exit(2)


def begin():
    # Set up logging before we do anything else
    if options.level_debug:
        options.level = 'DEBUG'
    utils.setup_root_logger(level=options.level, color=options.color)

    try:
        ioloop.IOLoop.instance().run_sync(main)
    except KeyboardInterrupt:
        log.info('CTRL-C Caught, shutting down')
    except Exception as e:
        # Skip traceback that involves tornado's libraries.
        import traceback
        trace_lines = traceback.format_exc(e).splitlines()
        skip_next = False
        for l in trace_lines:
            if 'tornado' in l:
                skip_next = True
                continue
            if not skip_next:
                print(l)
            skip_next = False

if __name__ == '__main__':
    begin()



import logging

from boto.exception import BotoServerError
from tornado import gen
from tornado import testing
import mock

from kingpin.actors.aws import settings
from kingpin.actors import exceptions
from kingpin.actors.aws import cloudformation

log = logging.getLogger(__name__)


@gen.coroutine
def tornado_value(*args):
    """Returns whatever is passed in. Used for testing."""
    raise gen.Return(*args)


class TestCloudFormationBaseActor(testing.AsyncTestCase):

    def setUp(self):
        super(TestCloudFormationBaseActor, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        reload(cloudformation)

    def test_init_with_bad_creds(self):
        settings.AWS_ACCESS_KEY_ID = None
        settings.AWS_SECRET_ACCESS_KEY = None
        with self.assertRaises(exceptions.InvalidCredentials):
            cloudformation.CloudFormationBaseActor(
                'unittest', {'region': 'us-east-1'})

    @testing.gen_test
    def test_get_stacks(self):
        actor = cloudformation.CloudFormationBaseActor(
            'unittest', {'region': 'us-east-1'})
        actor.cf_conn.list_stacks = mock.MagicMock()
        actor.cf_conn.list_stacks.return_value = [1, 2, 3]
        ret = yield actor._get_stacks()
        self.assertEquals([1, 2, 3], ret)

    @testing.gen_test
    def test_get_stack(self):
        actor = cloudformation.CloudFormationBaseActor(
            'unittest', {'region': 'us-east-1'})

        actor.cf_conn.list_stacks = mock.MagicMock()
        s1 = mock.MagicMock()
        s1.stack_name = 'stack-1'
        s2 = mock.MagicMock()
        s2.stack_name = 'stack-2'
        s3 = mock.MagicMock()
        s3.stack_name = 'stack-3'
        actor.cf_conn.list_stacks.return_value = [s1, s2, s3]

        ret = yield actor._get_stack('stack-1')
        self.assertEquals(ret, s1)

        ret = yield actor._get_stack('stack-5')
        self.assertEquals(ret, None)

    @testing.gen_test
    def test_wait_until_state(self):
        create_in_progress = mock.MagicMock(name='create_in_progress')
        create_in_progress.stack_status = 'CREATE_IN_PROGRESS'
        create_in_progress.stack_name = 'unittest'

        create_complete = mock.MagicMock(name='CREATE_COMPLETE')
        create_complete.stack_status = 'CREATE_COMPLETE'
        create_complete.stack_name = 'unittest'

        rollback_complete = mock.MagicMock(name='ROLLBACK_COMPLETE')
        rollback_complete.stack_status = 'ROLLBACK_COMPLETE'
        rollback_complete.stack_name = 'unittest'

        actor = cloudformation.CloudFormationBaseActor(
            'unittest', {'region': 'us-east-1'})
        actor._get_stack = mock.MagicMock()

        # Make _get_stack() yield back 2 in-progress states, then yield a
        # successfull execution.
        actor._get_stack.side_effect = [
            tornado_value(create_in_progress),
            tornado_value(create_in_progress),
            tornado_value(create_complete)
        ]
        ret = yield actor._wait_until_state(cloudformation.COMPLETE, sleep=0.1)
        self.assertEquals(ret, None)

        # Make sure a cloudformationerror is raised if we ask for a deleted
        # state rather than a created one.
        actor._get_stack.side_effect = [
            tornado_value(create_in_progress),
            tornado_value(create_in_progress),
            tornado_value(create_complete)
        ]
        with self.assertRaises(cloudformation.CloudFormationError):
            yield actor._wait_until_state(cloudformation.DELETED, sleep=0.1)

        # Lastly, test that if wait_until_state returns no actor, we bail
        # appropriately.
        actor._get_stack.side_effect = [tornado_value(None)]
        with self.assertRaises(cloudformation.StackNotFound):
            yield actor._wait_until_state(cloudformation.COMPLETE, sleep=0.1)


class TestCreate(testing.AsyncTestCase):

    def setUp(self):
        super(TestCreate, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        reload(cloudformation)

    def test_get_template_body(self):
        # Should work...
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template': 'examples/test/aws.cloudformation/cf.unittest.json'})
        # TODO: Fill this in with some real content
        self.assertNotEquals(actor._template_body, '#BLANK\r\n')
        self.assertEquals(actor._template_url, None)

        # Should return None
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template': 'http://foobar.json'})
        self.assertEquals(actor._template_body, None)
        self.assertEquals(actor._template_url, 'http://foobar.json')

        # Should raise exception
        with self.assertRaises(cloudformation.InvalidTemplate):
            actor = cloudformation.Create(
                'Unit Test Action',
                {'name': 'unit-test-cf',
                 'region': 'us-west-2',
                 'template': 'missing'})

    @testing.gen_test
    def test_validate_template_body(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template': 'examples/test/aws.cloudformation/cf.unittest.json'})
        actor.cf_conn.validate_template = mock.MagicMock()
        yield actor._validate_template()
        actor.cf_conn.validate_template.assert_called_with(
            template_body='#BLANK\n', template_url=None)

    @testing.gen_test
    def test_validate_template_url(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template': 'http://foobar.json'})
        actor.cf_conn.validate_template = mock.MagicMock()
        yield actor._validate_template()
        actor.cf_conn.validate_template.assert_called_with(
            template_body=None, template_url='http://foobar.json')

    @testing.gen_test
    def test_validate_template_raises_boto_error(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template': 'http://foobar.json'})
        actor.cf_conn.validate_template = mock.MagicMock()
        actor.cf_conn.validate_template.side_effect = BotoServerError(
            400, 'Invalid template property or properties')
        with self.assertRaises(cloudformation.InvalidTemplate):
            yield actor._validate_template()

        actor.cf_conn.validate_template.side_effect = BotoServerError(
            403, 'Invalid Credentials')
        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor._validate_template()

        actor.cf_conn.validate_template.side_effect = BotoServerError(
            500, 'Some other error')
        with self.assertRaises(BotoServerError):
            yield actor._validate_template()

    @testing.gen_test
    def test_create_stack(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'})
        actor.cf_conn.create_stack = mock.MagicMock(name='create_stack_mock')
        actor.cf_conn.create_stack.return_value = 'arn:123'
        ret = yield actor._create_stack()
        self.assertEquals(ret, 'arn:123')

    @testing.gen_test
    def test_create_stack_raises_boto_error(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'})
        actor.cf_conn.create_stack = mock.MagicMock()

        actor.cf_conn.create_stack.side_effect = BotoServerError(
            400, 'Invalid template property or properties')
        with self.assertRaises(cloudformation.CloudFormationError):
            yield actor._create_stack()

        actor.cf_conn.create_stack.side_effect = BotoServerError(
            403, 'Invalid credentials')
        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor._create_stack()

        actor.cf_conn.create_stack.side_effect = BotoServerError(
            500, 'Some unexpected failure')
        with self.assertRaises(BotoServerError):
            yield actor._create_stack()

    @testing.gen_test
    def test_execute(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'})

        actor._validate_template = mock.MagicMock()
        actor._validate_template.side_effect = [tornado_value(True)]

        actor._get_stack = mock.MagicMock()
        actor._get_stack.side_effect = [tornado_value(None)]

        actor._create_stack = mock.MagicMock()
        actor._create_stack.side_effect = [tornado_value(None)]

        actor._wait_until_state = mock.MagicMock()
        actor._wait_until_state.side_effect = [tornado_value(None)]
        yield actor._execute()

    @testing.gen_test
    def test_execute_exists(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'})

        actor._validate_template = mock.MagicMock()
        actor._validate_template.side_effect = [tornado_value(True)]

        actor._get_stack = mock.MagicMock()
        actor._get_stack.side_effect = [tornado_value(True)]

        with self.assertRaises(cloudformation.StackAlreadyExists):
            yield actor._execute()

    @testing.gen_test
    def test_execute_dry(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'},
            dry=True)

        actor._validate_template = mock.MagicMock()
        actor._validate_template.side_effect = [tornado_value(True)]

        actor._get_stack = mock.MagicMock()
        actor._get_stack.side_effect = [tornado_value(None)]

        yield actor._execute()


class TestDelete(testing.AsyncTestCase):

    def setUp(self):
        super(TestDelete, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        reload(cloudformation)

    @testing.gen_test
    def test_delete_stack(self):
        actor = cloudformation.Delete(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2'})
        actor.cf_conn.delete_stack = mock.MagicMock(name='delete_stack_mock')
        actor.cf_conn.delete_stack.return_value = 'req-id-1'
        ret = yield actor._delete_stack()
        self.assertEquals(ret, 'req-id-1')

    @testing.gen_test
    def test_delete_stack_raises_boto_error(self):
        actor = cloudformation.Delete(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2'})
        actor.cf_conn.delete_stack = mock.MagicMock()

        actor.cf_conn.delete_stack.side_effect = BotoServerError(
            400, 'Some error')
        with self.assertRaises(cloudformation.CloudFormationError):
            yield actor._delete_stack()

        actor.cf_conn.delete_stack.side_effect = BotoServerError(
            500, 'Some unexpected failure')
        with self.assertRaises(BotoServerError):
            yield actor._delete_stack()

    @testing.gen_test
    def test_execute(self):
        actor = cloudformation.Delete(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2'})
        actor._get_stack = mock.MagicMock()
        actor._get_stack.side_effect = [tornado_value(True)]
        actor._delete_stack = mock.MagicMock()
        actor._delete_stack.side_effect = [tornado_value(None)]
        actor._wait_until_state = mock.MagicMock()
        actor._wait_until_state.side_effect = cloudformation.StackNotFound()
        yield actor._execute()

    @testing.gen_test
    def test_execute_dry(self):
        actor = cloudformation.Delete(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2'}, dry=True)
        actor._get_stack = mock.MagicMock()
        actor._get_stack.side_effect = [tornado_value(True)]
        yield actor._execute()

    @testing.gen_test
    def test_execute_not_exists(self):
        actor = cloudformation.Delete(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2'})
        actor._get_stack = mock.MagicMock()
        actor._get_stack.side_effect = [tornado_value(None)]
        with self.assertRaises(cloudformation.StackNotFound):
            yield actor._execute()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2014 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.base`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The AWS Actors allow you to interact with the resources (such as SQS and ELB)
inside your Amazon AWS account. These actors all support dry runs properly, but
each actor has its own caveats with ``dry=True``. Please read the instructions
below for using each actor.

**Required Environment Variables**

:AWS_ACCESS_KEY_ID:
  Your AWS access key

:AWS_SECRET_ACCESS_KEY:
  Your AWS secret
"""

import logging
import re

from boto import utils as boto_utils
from boto import exception as boto_exception
from tornado import concurrent
from tornado import gen
from tornado import ioloop
from retrying import retry
import boto.cloudformation
import boto.ec2
import boto.ec2.elb
import boto.iam
import boto.sqs

from kingpin import utils
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.actors.aws import settings as aws_settings

log = logging.getLogger(__name__)

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'

EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class ELBNotFound(exceptions.RecoverableActorFailure):

    """Raised when an ELB is not found"""


class InvalidMetaData(exceptions.UnrecoverableActorFailure):

    """Raised when fetching AWS metadata."""


class AWSBaseActor(base.BaseActor):

    # Get references to existing objects that are used by the
    # tornado.concurrent.run_on_executor() decorator.
    ioloop = ioloop.IOLoop.current()
    executor = EXECUTOR

    all_options = {
        'region': (str, None, 'AWS Region (or zone) to connect to.')
    }

    def __init__(self, *args, **kwargs):
        """Check for required settings."""

        super(AWSBaseActor, self).__init__(*args, **kwargs)

        if not (aws_settings.AWS_ACCESS_KEY_ID and
                aws_settings.AWS_SECRET_ACCESS_KEY):
            raise exceptions.InvalidCredentials(
                'AWS settings imported but not all credentials are supplied. '
                'AWS_ACCESS_KEY_ID: %s, AWS_SECRET_ACCESS_KEY: %s' % (
                    aws_settings.AWS_ACCESS_KEY_ID,
                    aws_settings.AWS_SECRET_ACCESS_KEY))

        # Establish connection objects that don't require a region
        self.iam_conn = boto.iam.connection.IAMConnection(
            aws_access_key_id=aws_settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=aws_settings.AWS_SECRET_ACCESS_KEY)

        # Establish region-specific connection objects.
        region = self.option('region')
        if not region:
            return

        # In case a zone was provided instead of region we can convert
        # it on the fly
        zone_check = re.match(r'(.*[0-9])([a-z]*)$', region)

        if zone_check and zone_check.group(2):
            zone = region  # Only saving this for the log below

            # Set the fixed region
            region = zone_check.group(1)
            self.log.warning('Converting zone "%s" to region "%s".' % (
                zone, region))

        region_names = [r.name for r in boto.ec2.elb.regions()]
        if region not in region_names:
            err = ('Region "%s" not found. Available regions: %s' %
                   (region, region_names))
            raise exceptions.InvalidOptions(err)

        self.ec2_conn = boto.ec2.connect_to_region(
            region,
            aws_access_key_id=aws_settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=aws_settings.AWS_SECRET_ACCESS_KEY)

        self.elb_conn = boto.ec2.elb.connect_to_region(
            region,
            aws_access_key_id=aws_settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=aws_settings.AWS_SECRET_ACCESS_KEY)

        self.cf_conn = boto.cloudformation.connect_to_region(
            region,
            aws_access_key_id=aws_settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=aws_settings.AWS_SECRET_ACCESS_KEY)

        self.sqs_conn = boto.sqs.connect_to_region(
            region,
            aws_access_key_id=aws_settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=aws_settings.AWS_SECRET_ACCESS_KEY)

    @concurrent.run_on_executor
    @retry(**aws_settings.RETRYING_SETTINGS)
    @utils.exception_logger
    def thread(self, function, *args, **kwargs):
        """Execute `function` in a concurrent thread.

        Example:
            >>> zones = yield thread(ec2_conn.get_all_zones)

        This allows execution of any function in a thread without having
        to write a wrapper method that is decorated with run_on_executor()
        """
        try:
            return function(*args, **kwargs)
        except boto_exception.BotoServerError as e:
            msg = '%s: %s' % (e.error_code, e.message)

            if e.status == 403:
                raise exceptions.InvalidCredentials(msg)

            raise

    @gen.coroutine
    def _find_elb(self, name):
        """Return an ELB with the matching name.

        Must find exactly 1 match. Zones are limited by the AWS credentials.

        Args:
            name: String-name of the ELB to search for

        Returns:
            A single ELB reference object

        Raises:
            ELBNotFound
        """
        self.log.info('Searching for ELB "%s"' % name)

        try:
            elbs = yield self.thread(self.elb_conn.get_all_load_balancers,
                                     load_balancer_names=name)
        except boto_exception.BotoServerError as e:
            msg = '%s: %s' % (e.error_code, e.message)
            log.error('Received exception: %s' % msg)

            if e.status == 400:
                raise ELBNotFound(msg)

            raise

        self.log.debug('ELBs found: %s' % elbs)

        if len(elbs) != 1:
            raise ELBNotFound('Expected to find exactly 1 ELB. Found %s: %s'
                              % (len(elbs), elbs))

        raise gen.Return(elbs[0])

    @concurrent.run_on_executor
    @retry(**aws_settings.RETRYING_SETTINGS)
    def _get_meta_data(self, key):
        """Get AWS meta data for current instance.

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/
        ec2-instance-metadata.html
        """

        meta = boto_utils.get_instance_metadata(timeout=1, num_retries=2)
        if not meta:
            raise InvalidMetaData('No metadata available. Not AWS instance?')

        data = meta.get(key, None)
        if not data:
            raise InvalidMetaData('Metadata for key `%s` is not available')

        return data



"""Simple integration tests for the AWS IAM actors."""

from nose.plugins.attrib import attr
import logging

from tornado import testing

from kingpin.actors.aws import iam

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'

log = logging.getLogger(__name__)
logging.getLogger('boto').setLevel(logging.INFO)


class IntegrationIAM(testing.AsyncTestCase):

    integration = True

    cert_name = 'kingpin-integration-test'
    region = 'us-east-1'

    @attr('integration')
    @testing.gen_test(timeout=60)
    def integration_01a_upload_cert_dry(self):
        actor = iam.UploadCert(
            'Test',
            {'name': self.cert_name,
             'public_key_path': 'examples/test/server.pem',
             'private_key_path': 'examples/test/server.key'},
            dry=True)

        yield actor.execute()

    @attr('integration')
    @testing.gen_test(timeout=60)
    def integration_01b_upload_cert(self):
        actor = iam.UploadCert(
            'Test',
            {'name': self.cert_name,
             'public_key_path': 'examples/test/server.pem',
             'private_key_path': 'examples/test/server.key'})

        yield actor.execute()

    @attr('integration')
    @testing.gen_test(timeout=60)
    def integration_02a_delete_cert_dry(self):
        actor = iam.DeleteCert(
            'Test',
            {'name': self.cert_name},
            dry=True)

        yield actor.execute()

    @attr('integration')
    @testing.gen_test(timeout=60)
    def integration_02b_delete_cert(self):
        actor = iam.DeleteCert(
            'Test',
            {'name': self.cert_name})

        yield actor.execute()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2014 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.utils`
^^^^^^^^^^^^^^^^^^^^^^^^^^^

Misc methods for dealing with Actors.
"""

import logging

from kingpin import utils
from kingpin.actors import exceptions

log = logging.getLogger(__name__)


__author__ = 'Matt Wise <matt@nextdoor.com>'


def get_actor(config, dry):
    """Returns an initialized Actor object.

    Args:
        config: A dictionary of configuration data that conforms to our v1
                schema in kingpin.schema. Looks like this:

                {
                 'desc': <string description of actor>,
                 'actor': <string name of actor>
                 'options': <dict of options to pass to actor>
                 'warn_on_failure': <bool>
                 'condition': <string or bool>
                 }

        dry: Boolean whether or not in Dry mode
        warn_on_failure: Boolean

    Returns:
        <actor object>
    """
    # Copy the supplied dict before we modify it below
    config = dict(config)

    # Get the name of the actor, and pull it out of the config because its
    # not a valid kwarg for an Actor object.
    actor_string = config.pop('actor')

    log.debug('Building Actor "%s" with args: %s' % (actor_string, config))
    ActorClass = get_actor_class(actor_string)
    return ActorClass(dry=dry, **config)


def get_actor_class(actor):
    """Returns a Class Reference to an Actor by string name.

    Args:
        actor: String name of the actor to find.

    Returns:
        <Class Ref to Actor>
    """
    expected_exceptions = (AttributeError, ImportError, TypeError)

    try:
        # Try to load our local actors up first. Assume that the
        # 'kingpin.actors.' prefix was not included in the name.
        full_actor = 'kingpin.actors.%s' % actor
        ref = utils.str_to_class(full_actor)
    except expected_exceptions as e:
        log.warning('Could not import %s: %s' % (full_actor, e))
        try:
            ref = utils.str_to_class(actor)
        except expected_exceptions:
            log.critical('Could not import %s: %s' % (actor, e))
            msg = 'Unable to import "%s" as a valid Actor.' % actor
            raise exceptions.InvalidActor(msg)

    return ref



import os.path
import ConfigParser

CFG_USER_RC = '.rightscalerc'
CFG_SECTION_OAUTH = 'OAuth'
CFG_OPTION_ENDPOINT = 'api_endpoint'
CFG_OPTION_REF_TOKEN = 'refresh_token'

_config = None


class HookList(list):
    def __init__(self, *args, **kwargs):
        super(HookList, self).__init__(*args)
        for k, v in kwargs.items():
            setattr(self, k, v)


class HookDict(dict):
    pass


def get_config():
    global _config
    if not _config:
        _config = ConfigParser.SafeConfigParser()

        # set up some defaults - too bad only newer pythons know how to do this
        # more gracefully:
        _config.add_section(CFG_SECTION_OAUTH)
        _config.set(CFG_SECTION_OAUTH, CFG_OPTION_ENDPOINT, '')
        _config.set(CFG_SECTION_OAUTH, CFG_OPTION_REF_TOKEN, '')

        home = os.path.expanduser('~')
        rc_file = os.path.join(home, CFG_USER_RC)
        _config.read(rc_file)
    return _config


def get_rc_creds():
    """
    Reads ~/.rightscalerc and returns API endpoint and refresh token.

    Always returns a tuple of strings even if the file is empty - in which
    case, returns ``('', '')``.
    """
    config = get_config()
    try:
        return (
                config.get(CFG_SECTION_OAUTH, CFG_OPTION_ENDPOINT),
                config.get(CFG_SECTION_OAUTH, CFG_OPTION_REF_TOKEN),
                )
    except:
        return ('', '')


def find_by_name(collection, name, exact=True):
    """
    Searches collection by resource name.

    :param rightscale.ResourceCollection collection: The collection in which to
        look for :attr:`name`.

    :param str name: The name to look for in collection.

    :param bool exact: A RightScale ``index`` search with a :attr:`name` filter
        can return multiple results because it does a substring match on
        resource names.  So any resource that contains the specified name will
        be returned.  The :attr:`exact` flag controls whether to attempt to
        find an exact match for the given name.  If :attr:`exact` is ``False``,
        this will return a list of all the matches.  The default behaviour is
        to perform an exact match and return a single result.

    Returns ``None`` if no resources are found with a matching name.

    """
    params = {'filter[]': ['name==%s' % name]}
    found = collection.index(params=params)
    if not exact and len(found) > 0:
        return found
    for f in found:
        if f.soul['name'] == name:
            return f



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2014 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.mci`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. _MultiCloudImages_:
    http://reference.rightscale.com/api1.5/resources/ResourceMultiCloudImages.html
    http://reference.rightscale.com/api1.5/resources/ResourceMultiCloudImageSettings.html
"""

import logging

from tornado import gen

from kingpin.actors import exceptions
from kingpin.actors.rightscale import base
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


class MCIBaseActor(base.RightScaleBaseActor):

    """Abstract MCI Actor that provides some utility methods."""


class Create(MCIBaseActor):

    """Creates a RightScale Multi Cloud Image.

    Options match the documentation in RightScale:
    http://reference.rightscale.com/api1.5/resources/ResourceMultiCloudImages.html

    **Options**

    :name:
      The name of the MCI to be created.

    :description:
       The description of the MCI to be created.
       (*optional*)

    :images:
       A list of dicts that each describe a single cloud and the image in that
       cloud to launch. See below for details.

    **Image Definitions**

    Each cloud image definition is a dictionary that takes a few keys.

    :cloud:
      The name of the cloud as found in RightScale. We use the cloud 'Name'
      which can be found in your `Settings -> Account Settings -> Clouds ->
      insert_cloud_here` page. For example `AWS us-west-2`.

    :image:
      The cloud-specific Image UID. For example `ami-a1234abc`.

    :instance_type:
      The default instance type to launch when this AMI is launched. For
      example, `m1.small`.
      (*optional*)

    :user_data:
      The custom user data to pass to the instance on-bootup.
      (*optional*)

    **Examples**

    .. code-block:: json

        { "actor": "rightscale.mci.Create",
          "desc": "Create an MCI",
          "options": {
              "name": "Ubuntu i386 14.04",
              "description": "this is our test mci",
              "images": [
                {
                    "cloud": "EC2 us-west-2",
                    "image": "ami-e29774d1",
                    "instance_type": "m1.small",
                    "user_data": "cd /bin/bash"
                },
                {
                    "cloud": "EC2 us-west-1",
                    "image": "ami-b58142f1",
                    "instance_type": "m1.small",
                    "user_data": "cd /bin/bash"
                }
              ]
          }
        }

    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the MCI to be created.'),
        'description': (
            str, '',
            'The description of the MCI to be created.'),
        'images': (
            list, [],
            'A list of objects that describe our per cloud image settings.'),
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""

        super(Create, self).__init__(*args, **kwargs)

        allowed_image_options = (
            'cloud', 'image', 'instance_type', 'user_data')
        required_image_options = ('cloud', 'image')

        for image in self.option('images'):
            # Sanity check that no extra options were passed in
            for key in image.keys():
                if key not in allowed_image_options:
                    raise exceptions.InvalidOptions(
                        'Invalid option (%s) found in Image %s' % (key, image))

            # Make sure that the required options were passed in
            for required in required_image_options:
                if required not in image.keys():
                    raise exceptions.InvalidOptions(
                        'Missing option "%s" in Image %s' %
                        (required, image))

    @gen.coroutine
    def _get_image_def(self, description):
        """Returns a fully populated set of Multi Cloud Image settings.

        This method takes in a dictionary with as set of parameters (cloud,
        image, instance_type, user_data) and returns a fully ready-to-use
        set of RightScale-formatted parameters to create that image. The method
        handles discovering the RightScale HREFs for the cloud, image and
        instance_type options.

        Args:
            description: A dictionary with the keys: cloud, image,
            instance_type, user_data

        Returns:
            A RightScale-formatted array of tuples.
        """

        # Get our cloud object first -- its required so that we can search for
        # the image/ramdisk/etc hrefs.
        cloud = yield self._client.find_by_name_and_keys(
            collection=self._client._client.clouds,
            name=description['cloud'])
        if not cloud:
            raise exceptions.InvalidOptions(
                'Invalid Cloud name supplied: %s' % description['cloud'])

        # Find our image by searching for the resource_uid that matches.
        image = yield self._client.find_by_name_and_keys(
            collection=cloud.images,
            resource_uid=description['image'])
        if not image:
            raise exceptions.InvalidOptions(
                'Invalid cloud image name supplied: %s' % description['image'])

        # Find our instance type now too
        instance = yield self._client.find_by_name_and_keys(
            collection=cloud.instance_types,
            name=description['instance_type'])
        if not instance:
            raise exceptions.InvalidOptions(
                'Invalid cloud instance_type supplied: %s' %
                description['instance_type'])

        # Generate our mci parameters, and each of the image settings
        # parameters. This validates that our inputs are all correct one last
        # time.
        definition = self._generate_rightscale_params(
            prefix='multi_cloud_image_setting',
            params={
                'cloud_href': cloud.href,
                'image_href': image.href,
                'instance_type_href': instance.href,
                'user_data': description['user_data'],
            })

        self.log.debug('Prepared MCI Image Definition: %s' % definition)
        raise gen.Return(definition)

    @gen.coroutine
    def _execute(self):

        # Make sure the MCI doesn't already exist. If it does, we bail.
        mci = yield self._client.find_by_name_and_keys(
            collection=self._client._client.multi_cloud_images,
            name=self.option('name'))
        if mci:
            raise exceptions.InvalidOptions(
                'MCI "%s" already exists.' % self.option('name'))

        # Generate the parameters for creating the top level MCI object
        mci_params = self._generate_rightscale_params(
            prefix='multi_cloud_image',
            params={
                'description': self.option('description'),
                'name': self.option('name')
            })

        # Now, we need to validate that all of the inputs are correct by
        # discovering the hrefs for each of the images supplied.
        image_futures = []
        for image in self.option('images'):
            image_futures.append(self._get_image_def(image))
        mci_settings_params = yield image_futures

        # Finally, if we're dry, bail out..
        if self._dry:
            self.log.info('Would have created MCI: %s' % self.option('name'))
            for setting in mci_settings_params:
                self.log.info('Image Def: %s' % setting)
            raise gen.Return()

        # Ok, lets create this thing!
        self.log.info('Creating MCI %s' % self.option('name'))
        mci = yield self._client.create_resource(
            self._client._client.multi_cloud_images, mci_params)

        # Now add each of the image descriptions to the mci
        for setting in mci_settings_params:
            self.log.info('Creating MCI Setting: %s' % setting)
            yield self._client.create_resource(
                mci.settings, setting)


class Destroy(MCIBaseActor):

    """Deletes a RightScale MCI.

    Options match the documentation in RightScale:
    http://reference.rightscale.com/api1.5/resources/ResourceMultiCloudImages.html

    **Options**

    :name:
      The name of the multi cloud image to be deleted.

    **Examples**

    .. code-block:: json

        { "actor": "rightscale.mci.Destroy",
          "desc": "Create an MCI",
          "options": {
              "name": "Ubuntu i386 14.04",
          }
        }

    """

    all_options = {
        'name': (str, REQUIRED,
                 'The name of the multi cloud image to be deleted.'),
    }

    @gen.coroutine
    def _execute(self):

        mci = yield self._client.find_by_name_and_keys(
            collection=self._client._client.multi_cloud_images,
            name=self.option('name'))
        if not mci:
            raise exceptions.InvalidOptions(
                'MCI "%s" does not exist.' % self.option('name'))

        info = (yield self._client.show(mci.self)).soul

        if self._dry:
            self.log.info('Would delete MCI %s' % info['name'])
            raise gen.Return()

        self.log.info('Deleting MCI %s' % info['name'])
        yield self._client.destroy_resource(mci)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2014 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.alerts`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import logging

from tornado import gen
import requests

from kingpin.actors import exceptions
from kingpin.actors.rightscale import base
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


class InvalidInputs(exceptions.InvalidOptions):

    """Raised when supplied inputs are invalid for a ServerArray."""


class AlertSpecNotFound(exceptions.RecoverableActorFailure):

    """Raised when an Alert Spec could not be found"""


class CreateFailed(exceptions.RecoverableActorFailure):

    """Raised when an Alert Spec could not be created"""


class AlertsBaseActor(base.RightScaleBaseActor):

    """Abstract Alerts Actor that provides some utility methods."""

    @gen.coroutine
    def _find_alert_spec(self, name, subject_href):
        """Search for an AlertSpec by-name and return the resource.

        Note: A non-exact resource match is used below so that we return all of
        the Alert Specs that are matched by name. This method returns the
        resources in a list.

        Args:
            name: RightScale AlertSpec Name
            subject_href: The HREF of the subject this AlertSpec is assigned
              to.

        Return:
            [<rightcale.Resource objects>]
        """
        log.debug('Searching for AlertSpec matching: %s' % name)
        found_spec = yield self._client.find_by_name_and_keys(
            self._client._client.alert_specs, name, exact=False,
            subject_href=subject_href)

        if not found_spec:
            log.debug('AlertSpec matching "%s" could not be found.' % name)
            return

        log.debug('Got AlertSpec: %s' % found_spec)
        raise gen.Return(found_spec)


class Create(AlertsBaseActor):

    """Create a RightScale Alert Spec

    Options match the documentation in RightScale:
    http://reference.rightscale.com/api1.5/resources/ResourceAlertSpecs.html#create

    **Options**

    :array:
      The name of the Server or ServerArray to create the AlertSpec on.

    :strict_array:
      Whether or not to fail if the Server/ServerArray does not exist.
      (default: False)

    :condition:
      The condition (operator) in the condition sentence.
      (`>, >=, <, <=, ==, !=`)

    :description:
      The description of the AlertSpec.
      (*optional*)

    :duration:
      The duration in minutes of the condition sentence.
      (`^\d+$`)

    :escalation_name:
      Escalate to the named alert escalation when the alert is triggered.
      (*optional*)

    :file:
      The RRD path/file_name of the condition sentence.

    :name:
      The name of the AlertSpec.

    :threshold:
      The threshold of the condition sentence.

    :variable:
      The RRD variable of the condition sentence

    :vote_tag:
      Should correspond to a vote tag on a ServerArray if vote to grow or
      shrink.

    :vote_type:
      Vote to grow or shrink a ServerArray when the alert is triggered. Must
      either escalate or vote.
      (`grow` or `shrink`)

    **Examples**

    Create a high network activity alert on my-array:

    .. code-block:: json

       { "desc": "Create high network rx alert",
         "actor": "rightscale.alerts.Create",
         "options": {
           "array": "my-array",
           "strict_array": true,
           "condition": ">",
           "description": "Alert if amount of network data received is high",
           "duration": 180,
           "escalation_name": "Email Engineering",
           "file": "interface/if_octets-eth0",
           "name": "high network rx activity",
           "threshold": "50000000",
           "variable": "rx"
         }
       }

    **Dry Mode**

    In Dry mode this actor *does* validate that the ``array`` array exists.
    If it does not, a `kingpin.actors.rightscale.api.ServerArrayException` is
    thrown. Once that has been validated, the dry mode execution simply logs
    the Alert Spec that it would have created.

    Example *dry* output::

        TODO: Fill this in
    """

    all_options = {
        'array': (str, REQUIRED, 'Name of the ServerArray act on.'),
        'strict_array': (bool, False,
                         ('Whether or not to fail if the  Server/ServerArray ',
                          'does not exist.')),
        'condition': (str, REQUIRED,
                      'The condition (operator) in the condition sentence.'),
        'description': (str, None, 'The description of the AlertSpec.'),
        'duration': ((int, str), REQUIRED,
                     'The duration in minutes of the condition sentence.'),
        'escalation_name': (str, None,
                            ('Escalate to the named alert escalation when the',
                             'alert is triggered. Must either escalate or',
                             'vote.')),
        'file': (str, REQUIRED,
                 'The RRD path/file_name of the condition sentence.'),
        'name': (str, REQUIRED, 'The name of the AlertSpec.'),
        'threshold': (str, REQUIRED,
                      'The threshold of the condition sentence.'),
        'variable': (str, REQUIRED,
                     'The RRD variable of the condition sentence.'),
        'vote_tag': (str, None,
                     ('Should correspond to a vote tag on a ServerArray if ',
                      'vote to grow or shrink.')),
        'vote_type': (str, None,
                      ('Vote to grow or shrink a ServerArray when the alert ',
                       'is triggered. Must either escalate or vote.'))
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""
        super(Create, self).__init__(*args, **kwargs)
        # By default, we're strict on our array array validation
        self._array_raise_on = 'notfound'
        self._array_allow_mock = False

        if not self.option('strict_array'):
            self._array_raise_on = None
            self._array_allow_mock = True

        if self.option('vote_type') not in ('grow', 'shrink', None):
            raise exceptions.InvalidOptions(
                'vote_type must be either: grow, shrink, None')

    @gen.coroutine
    def _execute(self):
        # Find the array we're adding an alert spec to. Specifically, we need
        # the servers HREF.
        array = yield self._find_server_arrays(
            self.option('array'),
            raise_on=self._array_raise_on,
            allow_mock=self._array_allow_mock)
        self.log.info('Found %s (%s)' % (array.soul['name'], array.href))

        # Add all of the required parameters to a dictionary
        params = {
            'condition': self.option('condition'),
            'description': self.option('description'),
            'duration': int(self.option('duration')),
            'file': self.option('file'),
            'name': self.option('name'),
            'subject_href': array.href,
            'threshold': self.option('threshold'),
            'variable': self.option('variable'),
        }

        # Generate the RightScale parameters that we need to pass in when
        # creating the alert. The optional parameters should not be passed in
        # if their option value came in as None.
        _optional_params = [
            'description', 'escalation_name', 'vote_tag', 'vote_type'
        ]
        for optional in _optional_params:
            if self.option(optional):
                params[optional] = self.option(optional)

        params = self._generate_rightscale_params('alert_spec', params)
        self.log.debug('Generated params: %s' % params)

        if self._dry:
            # In dry run mode, just log out what we would have done.
            self.log.info('Would have created the alert spec \"%s\" on %s' %
                          (self.option('name'), array.soul['name']))
            raise gen.Return()

        # We're really doin this. If we get a known exception back, handle
        # it. Otherwise, raise it.
        try:
            yield self._client.create_resource(
                self._client._client.alert_specs, params)
            self.log.info('Alert spec has been created')
        except requests.exceptions.HTTPError as e:
            if e.response.status_code in (422, 400):
                msg = ('Invalid parameters supplied to Alert Spec "%s": %s'
                       % (self.option('name'), params))
                raise exceptions.RecoverableActorFailure(msg)
            raise


class Destroy(AlertsBaseActor):

    """Destroy existing RightScale Alert Specs

    This actor searches RightScale for any Alert Specs that match the ``name``
    and ``array`` that you supplied, then deletes all of them. RightScale lets
    you have multiple alert specs with the same name, so if this actor finds
    multiple specs, it will delete them all.

    **Options**

    :array:
      The name of the Server or ServerArray to delete the AlertSpec from.

    :name:
      The name of the AlertSpec.

    **Examples**

    Destroy a high network activity alert on my-array:

    .. code-block:: json

       { "desc": "Destroy high network rx alert",
         "actor": "rightscale.alerts.Destroy",
         "options": {
           "array": "my-array",
           "name": "high network rx activity",
         }
       }

    **Dry Mode**

    In Dry mode this actor *does* validate that the ``array`` array exists,
    and that the AlertSpec exists on that array so that it can be deleted. A
    RecoverableActorFailure error is thrown if it does not exist.

    Example *dry* output::

        14:31:49   INFO      Rehearsing... Break a leg!
        14:31:49   INFO      [DRY: Kingpin] Preparing actors from delete.json
        14:31:53   INFO      [DRY: Destroy high network rx alert] Found
          my-array (/api/server_arrays/329142003) to delete alert spec from
        14:31:54   INFO      [DRY: Destroy high network rx alert] Would have
          deleted the alert spec "high network rx activity" on my-array
    """

    all_options = {
        'array': (str, REQUIRED, 'Name of the ServerArray act on.'),
        'name': (str, REQUIRED, 'The name of the AlertSpec.')
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""
        super(Destroy, self).__init__(*args, **kwargs)
        # By default, we're strict on our array validation
        self._array_raise_on = 'notfound'
        self._array_allow_mock = False

    @gen.coroutine
    def _execute(self):
        # Find the array we're adding an alert spec to. Specifically, we need
        # the servers HREF.
        array = yield self._find_server_arrays(
            self.option('array'),
            raise_on=self._array_raise_on,
            allow_mock=self._array_allow_mock)

        self.log.info('Found %s (%s) to delete alert spec from' %
                      (array.soul['name'], array.href))

        # Find the AlertSpec on this server, if it exists.
        alerts = yield self._find_alert_spec(
            self.option('name'), array.href)

        # If we can't find the AlertSpec specific to the subjet array that was
        # supplied, raise an exception and bail.
        if not alerts:
            raise AlertSpecNotFound(
                '"%s" could not be found on %s' %
                (self.option('name'), array.soul['name']))

        # We'll store our 'delete spec' futures in here
        deletes = []

        for spec in alerts:
            log.debug('Found Alert Spec %s' % spec.soul)

            if self._dry:
                # In dry run mode, just log out what we would have done.
                self.log.info('Would have deleted alert \"%s\" (%s) on %s' %
                              (spec.soul['name'],
                               spec.href,
                               array.soul['name']))
            else:
                # We're really doin this!
                self.log.info('Deleting alert \"%s\" (%s) on %s' %
                              (spec.soul['name'],
                               spec.href,
                               array.soul['name']))
                deletes.append(self._client.destroy_resource(spec))

        # Wait for the deletes to finish
        if deletes:
            yield deletes



"""Vanity test"""

import unittest


class TestVersion(unittest.TestCase):

    def test_version(self):
        from kingpin import version
        reload(version)
        self.assertEquals(type(version.__version__), str)
